#!/usr/bin/env python3
#
# ############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import os
import shutil
import subprocess
import json
import pysam
import gzip
import sys

from .common import get_path_to_program
from .error_codes import IsoQuantExitCode
from .gtf2db import convert_db_to_gtf, db2bed
from .input_data_storage import SampleData, InputDataType

logger = logging.getLogger('IsoQuant')

PACBIO_CCS_DATA = 'pacbio_ccs'
NANOPORE_DATA = 'nanopore'
ASSEMBLY = 'assembly'

DATA_TYPE_ALIASES = {PACBIO_CCS_DATA: PACBIO_CCS_DATA, "pacbio": PACBIO_CCS_DATA, NANOPORE_DATA: NANOPORE_DATA,
                     "ont": NANOPORE_DATA, ASSEMBLY: ASSEMBLY, "transcripts": ASSEMBLY}
DATATYPE_TO_ALIGNER = {ASSEMBLY: 'minimap2', PACBIO_CCS_DATA: 'minimap2', NANOPORE_DATA: 'minimap2'}
                       # 'barcoded_se_reads' : 'star', 'barcoded_pe_reads' : 'star'}

SUPPORTED_ALIGNERS = ['starlong', 'minimap2']
SUPPORTED_STRANDEDNESS = ['forward', 'reverse', 'none']

KMER_SIZE = {ASSEMBLY: '15', PACBIO_CCS_DATA: '15', NANOPORE_DATA: '14'}
# CANONICAL_SITE_BONUS = {ASSEMBLY: '5', PACBIO_CCS_DATA: '5', NANOPORE_DATA: '9'}
MINIMAP_PRESET = {ASSEMBLY: 'splice:hq', PACBIO_CCS_DATA: 'splice:hq', NANOPORE_DATA: 'splice'}

class DataSetReadMapper:
    def __init__(self, args):
        self.args = args
        self.aligner = self.choose_aligner()
        self.index_fname = self.create_index(args)

    def choose_aligner(self):
        return self.args.aligner or DATATYPE_TO_ALIGNER[self.args.data_type]

    def create_index(self, args):
        if args.index and os.path.exists(args.index):
            return args.index

        index = None if args.clean_start else find_stored_index(args, self.aligner)

        if index is None:
            index = index_reference(self.aligner, args)
            store_index(index, args, self.aligner)
        return index

    def map_reads(self, args):
        samples = []
        annotation_file = find_annotation(self.aligner, args)
        for sample in args.input_data.samples:
            readable_names_dict = {}
            bam_files = []
            for fastq_files in sample.file_list:
                for fastq_file in fastq_files:
                    bam_file = None if args.clean_start else find_stored_alignment(fastq_file, annotation_file, args)
                    if bam_file is None:
                        bam_file = align_reads(self.aligner, fastq_file, annotation_file, args, sample.prefix, sample.aux_dir, args.input_data.input_type)
                        store_alignment(bam_file, fastq_file, annotation_file, args)
                    bam_files.append([bam_file])
                    if fastq_file in sample.readable_names_dict:
                        readable_names_dict[bam_file] = sample.readable_names_dict[fastq_file]
            samples.append(SampleData(bam_files, sample.prefix, sample.out_dir, readable_names_dict, sample.illumina_bam, sample.barcoded_reads))

        args.input_data.samples = samples
        args.input_data.input_type = InputDataType.bam
        return args.input_data


def get_aligner(aligner):
    path = get_path_to_program(aligner)
    if not path:
        logger.critical('{aligner} is not found! Make sure that {aligner} is in your PATH or use other alignment method'.format(aligner=aligner))
        sys.exit(IsoQuantExitCode.ALIGNER_NOT_FOUND)
    return path


def load_json(json_path):
    if not os.path.isfile(json_path):
        return {}

    try:
        with open(json_path, "r") as json_file:
            res_dict = json.load(json_file)
    except json.decoder.JSONDecodeError:
        logger.error('JSON config {json_path} was read with errors, overwriting now'.format(json_path=json_path))
        with open(json_path, "w") as json_file:
            json.dump({}, json_file, indent=2)
        res_dict = {}

    return res_dict


def find_stored_index(args, aligner):
    reference_filename = os.path.abspath(args.reference)
    converted_indexes = load_json(args.index_config_path)

    index_filename = converted_indexes.get(reference_filename, {}).get('index_filename')
    logger.debug('Searching for previously created index for {}'.format(reference_filename))
    if index_filename is None:
        return None
    index_mtime = converted_indexes.get(reference_filename, {}).get('index_mtime')
    reference_mtime = converted_indexes.get(reference_filename, {}).get('reference_mtime')
    used_aligner = converted_indexes.get(reference_filename, {}).get('aligner')
    if os.path.exists(reference_filename) and os.path.getmtime(reference_filename) == reference_mtime:
        if os.path.exists(index_filename) and os.path.getmtime(index_filename) == index_mtime:
            if aligner + str(KMER_SIZE[args.data_type]) == used_aligner:
                logger.info('Index file found. Using {}'.format(index_filename))
                return index_filename
    return None


def store_index(index, args, aligner):
    reference_filename = os.path.abspath(args.reference)
    index = os.path.abspath(index)

    converted_indexes = load_json(args.index_config_path)
    converted_indexes[reference_filename] = {
        'index_filename': index,
        'reference_mtime': os.path.getmtime(reference_filename),
        'index_mtime': os.path.getmtime(index),
        'aligner': aligner + str(KMER_SIZE[args.data_type])
    }
    with open(args.index_config_path, 'w') as f_out:
        json.dump(converted_indexes, f_out, indent=2)
    logger.debug('New index saved to {}'.format(index))


def find_stored_bed(args):
    genedb_filename = os.path.abspath(args.genedb)
    converted_beds = load_json(args.bed_config_path)

    bed_filename = converted_beds.get(genedb_filename, {}).get('bed_filename')
    logger.debug('Searching for previously created BED for {}'.format(genedb_filename))
    if bed_filename is None:
        return None
    bed_mtime = converted_beds.get(genedb_filename, {}).get('bed_mtime')
    reference_mtime = converted_beds.get(genedb_filename, {}).get('reference_mtime')
    if os.path.exists(genedb_filename) and os.path.getmtime(genedb_filename) == reference_mtime:
        if os.path.exists(bed_filename) and os.path.getmtime(bed_filename) == bed_mtime:
                logger.info('BED file found. Using {}'.format(bed_filename))
                return bed_filename
    return None


def store_bed(bed, args):
    genedb_filename = os.path.abspath(args.genedb)
    bed = os.path.abspath(bed)

    converted_beds = load_json(args.bed_config_path)
    converted_beds[genedb_filename] = {
        'bed_filename': bed,
        'reference_mtime': os.path.getmtime(genedb_filename),
        'bed_mtime': os.path.getmtime(bed)
    }
    with open(args.bed_config_path, 'w') as f_out:
        json.dump(converted_beds, f_out, indent=2)
    logger.debug('New BED saved to {}'.format(bed))


def find_stored_alignment(fastq_file, annotation, args):
    fastq = os.path.abspath(fastq_file)
    index = os.path.abspath(args.index)
    ann_path = os.path.abspath(annotation) if annotation else ""
    ann_str = "_" + ann_path if ann_path else ""

    key = "%s_aligned_to_%s%s" % (fastq, index, ann_str)
    aligned_fastq_files = load_json(args.alignment_config_path)

    logger.debug('Searching for previously created alignment for {}'.format(fastq))
    bam_fpath = aligned_fastq_files.get(key, {}).get('alignment_fpath')
    if bam_fpath is None:
        return None

    index_mtime = aligned_fastq_files.get(key, {}).get('index_mtime')
    if os.path.getmtime(index) != index_mtime:
        return None
    if ann_path:
        ann_mtime = aligned_fastq_files.get(key, {}).get('ann_mtime')
        if os.path.getmtime(ann_path) != ann_mtime:
            return None

    fastq_mtime = aligned_fastq_files.get(key, {}).get('fastq_mtime')
    bam_mtime = aligned_fastq_files.get(key, {}).get('bam_mtime')
    if os.path.exists(fastq) and os.path.getmtime(fastq) == fastq_mtime:
        if os.path.exists(bam_fpath) and os.path.getmtime(bam_fpath) == bam_mtime:
            logger.info('Bam alignment file found. Using {}'.format(bam_fpath))
            return bam_fpath
    return None


def store_alignment(bam_file, fastq_file, annotation, args):
    fastq = os.path.abspath(fastq_file)
    index = os.path.abspath(args.index)
    ann_path = os.path.abspath(annotation) if annotation else ""

    key = "%s_aligned_to_%s%s" % (fastq, index, "_" + ann_path if ann_path else "")
    bam_file = os.path.abspath(bam_file)
    aligned_fastq_files = load_json(args.alignment_config_path)
    aligned_fastq_files[key] = {
        'alignment_fpath': bam_file,
        'index_mtime': os.path.getmtime(index),
        'fastq_mtime': os.path.getmtime(fastq),
        'bam_mtime': os.path.getmtime(bam_file),
        'ann_mtime': os.path.getmtime(ann_path) if ann_path else ""
    }
    with open(args.alignment_config_path, 'w') as f_out:
        json.dump(aligned_fastq_files, f_out, indent=2)
    logger.debug('New alignment saved to {}'.format(bam_file))


def index_reference(aligner, args):
    logger.info('Indexing reference')
    ref_name, _ = os.path.splitext(args.reference.split('/')[-1])

    command = ""
    index_name = os.path.join(os.path.abspath(args.output), "%s_k%s_idx" % (ref_name, KMER_SIZE[args.data_type]))

    if aligner == "starlong":
        if os.path.isdir(index_name) and os.path.exists(os.path.join(index_name, "genomeParameters.txt")):
            logger.debug('Reusing reference index ' + index_name)
            return index_name
        exec_path = get_aligner('STARlong')
        if not os.path.isdir(index_name):
            os.makedirs(index_name)
        command = [exec_path, '--runMode', 'genomeGenerate', '--runThreadN', str(args.threads),
                   '--genomeDir', index_name, '--genomeFastaFiles', args.reference]
        if args.indexing_options:
            command += args.indexing_options.split()

    elif aligner == "minimap2":
        if os.path.isfile(index_name):
            logger.debug('Reusing reference index ' + index_name)
            return index_name
        minimap2_path = get_aligner('minimap2')
        command = [minimap2_path, '-t', str(args.threads),
                   '-k', str(KMER_SIZE[args.data_type]),
                   '-w', '5']

        if args.indexing_options:
            command += args.indexing_options.split()
        command += ['-d', index_name, args.reference]

    else:
        logger.critical("Aligner " + aligner + " is not supported")
        sys.exit(IsoQuantExitCode.ALIGNER_NOT_FOUND)

    log_fpath = os.path.join(args.output, "alignment.log")
    log_file = open(log_fpath, "w")
    if subprocess.call(command, stdout=log_file, stderr=log_file) != 0:
        logger.critical("Failed indexing reference! See " + log_fpath)
        sys.exit(IsoQuantExitCode.INDEXING_FAILED)
    return index_name


def find_annotation(aligner, args):
    if args.no_junc_bed:
        return None

    if aligner == "starlong":
        gene_annotation = args.genedb
        if gene_annotation.lower().endswith("db"):
            if args.original_annotation:
                gene_annotation = args.original_annotation
            else:
                gene_annotation = os.path.abspath(convert_db_to_gtf(args))

        if gene_annotation.lower().endswith("gz") or gene_annotation.lower().endswith("gzip"):
            genedb_file_name = os.path.basename(gene_annotation)
            genedb_name, _ = os.path.splitext(genedb_file_name)
            gunzipped_genedb = os.path.join(args.output, genedb_name)
            if not os.path.exists(gunzipped_genedb) or not args.resume:
                logger.info("Decompressing genome annotation to " + str(gunzipped_genedb))
                with open(gunzipped_genedb, "w") as outf:
                    shutil.copyfileobj(gzip.open(gene_annotation, "rt"), outf)
            gene_annotation = os.path.abspath(gunzipped_genedb)

        return os.path.abspath(gene_annotation)

    elif aligner == "minimap2":
        bed_fname = None
        if args.junc_bed_file:
            if not os.path.exists(args.junc_bed_file):
                logger.warning("Provided BED file %s does not exist, will generate from the annotation" %
                               args.junc_bed_file)
            else:
                bed_fname = args.junc_bed_file

        if bed_fname is None:
            bed_fname = find_stored_bed(args)
            if bed_fname is None:
                bed_fname = os.path.join(args.output, os.path.splitext(os.path.basename(args.genedb))[0] + ".bed")
                db2bed(args.genedb, bed_fname)
                store_bed(bed_fname, args)

        return os.path.abspath(bed_fname)


def align_reads(aligner, reads_file, annotation_file, args, label, out_dir, data_type=InputDataType.fastq):
    reads_path = os.path.abspath(reads_file)
    fname, ext = os.path.splitext(reads_path.split('/')[-1])
    alignment_prefix = str(os.path.join(out_dir, label))

    prefix_name = fname
    if prefix_name.endswith(".fq") or prefix_name.endswith(".fastq") or prefix_name.endswith(".fa") or prefix_name.endswith(".fasta") or prefix_name.endswith(".bam"):
        prefix_name, _ = os.path.splitext(prefix_name)
    hash_annotation = "_" + ("%x" % hash(annotation_file))[2:8] if annotation_file else ""
    hash_index = ("%x" % hash(args.index))[2:8]
    hash_fastq = ("%x" % hash(reads_path))[2:8]

    alignment_bam_path = str(os.path.join(out_dir, label + '_' + prefix_name + '_%s_%s%s.bam' % (hash_fastq, hash_index, hash_annotation)))
    logger.info("Aligning %s to the reference, alignments will be saved to %s" % (os.path.abspath(reads_path),
                                                                                  os.path.abspath(alignment_bam_path)))
    alignment_sam_path = alignment_bam_path[:-4] + '.sam'

    log_fpath = os.path.join(args.output, "alignment.log")
    if aligner == "starlong":
        if data_type != InputDataType.fastq:
            logger.critical("Input data type must be FASTA/FASTQ when using star aligner, unmapped BAM is not supported yet")
            sys.exit(IsoQuantExitCode.INVALID_PARAMETER)
        star_path = get_aligner('STARlong')
        zcat_option = ['--readFilesCommand', 'zcat'] if (ext.endswith('gz') or ext.endswith('gzip')) else []
        # command = '{star} --runThreadN 16 --genomeDir {ref_index_name}  --readFilesIn {transcripts}  --outSAMtype SAM
        #  --outFileNamePrefix {alignment_out}'.format(star=star_path, ref_index_name=star_index, transcripts=short_id_contigs_name, alignment_out=alignment_sam_path)
        annotation_opts = [] if not annotation_file else ['--sjdbGTFfile', annotation_file, '--sjdbOverhang', '140']
        command = ([star_path, '--runThreadN', str(args.threads), '--genomeDir', args.index, '--readFilesIn',
                   reads_path, '--outSAMtype', 'BAM', 'SortedByCoordinate', '--seedPerReadNmax', '1000000',
                   '--outBAMsortingThreadN', str(args.threads), '--outSAMattributes', 'NH', 'HI', 'NM', 'MD'] +
                   annotation_opts + zcat_option + ['--outFileNamePrefix', alignment_prefix])
        if args.mapping_options:
            command += args.mapping_options.split()

        logger.info("Running STAR (takes a while)")
        with open(log_fpath, "a") as log_file:
            if subprocess.call(command, stdout=log_file, stderr=log_file) != 0:
                logger.critical("STAR finished with errors! See " + log_fpath)
                sys.exit(IsoQuantExitCode.ALIGNMENT_FAILED)
        shutil.move(alignment_prefix + "Aligned.sortedByCoord.out.bam", alignment_bam_path)

    elif aligner == "minimap2":
        additional_options = []
        minimap2_path = get_aligner('minimap2')
        if args.stranded == 'forward':
            additional_options.append('-uf')
        if annotation_file:
            args.junc_bed_file = annotation_file
            additional_options.append("--junc-bed")
            additional_options.append(annotation_file)

        if data_type == InputDataType.fastq:
            command = [minimap2_path, args.index, reads_path, '-a', '-x', MINIMAP_PRESET[args.data_type],
                       '--secondary=yes', '-Y', '--MD', '-t', str(args.threads)] + additional_options
        elif data_type == InputDataType.unmapped_bam:
            command = [minimap2_path, args.index, '-', '-a', '-x', MINIMAP_PRESET[args.data_type],
                       '--secondary=yes', '-Y', '--MD', '-t', str(args.threads)] + additional_options
        else:
            logger.critical("Data type %s is not compatible with minimap2" % data_type.name)
            sys.exit(IsoQuantExitCode.INVALID_PARAMETER)

        if args.mapping_options:
            command += args.mapping_options.split()

        minimap_version = "unknown"
        version_run = subprocess.run([minimap2_path, '--version'], capture_output=True)
        if version_run.returncode == 0:
            minimap_version = version_run.stdout.decode('UTF-8').strip()

        logger.info("Running minimap2 version %s (takes a while)" % minimap_version)
        if data_type == InputDataType.fastq:
            with open(log_fpath, "a") as log_file:
                minimap2_process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=log_file)
                samtools_sort_process = subprocess.Popen(['samtools', 'sort', '-@', str(args.threads), '-o', alignment_bam_path],
                                                         stdin=minimap2_process.stdout, stderr=log_file)
                minimap2_return_code = minimap2_process.wait()
                minimap2_process.stdout.close()
                if minimap2_return_code != 0:
                    logger.critical("Minimap2 finished with errors! See " + log_fpath)
                    sys.exit(IsoQuantExitCode.ALIGNMENT_FAILED)
                samtools_return_code = samtools_sort_process.wait()
                if samtools_return_code != 0:
                    logger.critical("Samtools sort finished with errors! See " + log_fpath)
                    sys.exit(IsoQuantExitCode.SAMTOOLS_FAILED)

        elif data_type == InputDataType.unmapped_bam:
            with open(log_fpath, "a") as log_file:
                bam_open_process = subprocess.Popen(['samtools', 'fastq', reads_path], stdout=subprocess.PIPE, stderr=log_file)
                minimap2_process = subprocess.Popen(command, stdin=bam_open_process.stdout, stdout=subprocess.PIPE, stderr=log_file)
                samtools_sort_process = subprocess.Popen(['samtools', 'sort', '-@', str(args.threads), '-o', alignment_bam_path],
                                                         stdin=minimap2_process.stdout, stderr=log_file)

                bam_open_return_code = bam_open_process.wait()
                bam_open_process.stdout.close()
                if bam_open_return_code != 0:
                    logger.critical("BAM to FASTQ conversion finished with errors! See " + log_fpath)
                    sys.exit(IsoQuantExitCode.SUBPROCESS_FAILED)
                minimap2_return_code = minimap2_process.wait()
                minimap2_process.stdout.close()
                if minimap2_return_code != 0:
                    logger.critical("Minimap2 finished with errors! See " + log_fpath)
                    sys.exit(IsoQuantExitCode.ALIGNMENT_FAILED)
                samtools_return_code = samtools_sort_process.wait()
                if samtools_return_code != 0:
                    logger.critical("Samtools sort finished with errors! See " + log_fpath)
                    sys.exit(IsoQuantExitCode.SAMTOOLS_FAILED)

    else:
        logger.critical("Aligner " + aligner + " is not supported")
        sys.exit(IsoQuantExitCode.ALIGNER_NOT_FOUND)
    logger.info("Indexing alignments")
    try:
        pysam.index(alignment_bam_path)
    except pysam.SamtoolsError as err:
        if "failed to create index" in err.value:
            logger.info(f"Samtools failed to generate default .bai index; with error: {err.value}")
            logger.info("Trying to build a CSI index instead")
            try:
                pysam.index('-@', str(args.threads), '-c', alignment_bam_path)
            except pysam.SamtoolsError as err:
                logger.error(f"Failed to create CSI index: {err.value}")
                sys.exit(IsoQuantExitCode.SAMTOOLS_FAILED)
        else:
            logger.error(err.value)
            sys.exit(IsoQuantExitCode.SAMTOOLS_FAILED)
    except OSError as err:
        logger.error(err)
        sys.exit(IsoQuantExitCode.SUBPROCESS_FAILED)

    return alignment_bam_path




############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pytest
import tempfile
import os
from src.read_groups import (
    AlignmentTagReadGrouper,
    ReadIdSplitReadGrouper,
    ReadTableGrouper,
)


# Mock alignment object
class MockAlignment:
    def __init__(self, read_id="", tags=None):
        self.query_name = read_id
        if tags is None:
            tags = {}
        self._tags = tags

    def get_tag(self, tag_name):
        return self._tags.get(tag_name)

    def has_tag(self, tag_name):
        return tag_name in self._tags


class TestAlignmentTagReadGrouper:
    """Test AlignmentTagReadGrouper class."""

    def test_init_default_tag(self):
        """Test AlignmentTagReadGrouper with default RG tag."""
        grouper = AlignmentTagReadGrouper()
        assert grouper.tag == "RG"

    def test_init_custom_tag(self):
        """Test AlignmentTagReadGrouper with custom tag."""
        grouper = AlignmentTagReadGrouper("CB")
        assert grouper.tag == "CB"

    def test_get_group_with_tag(self):
        """Test getting group when read has tag."""
        grouper = AlignmentTagReadGrouper("CB")
        alignment = MockAlignment(tags={"CB": "ACTGACTG"})
        assert grouper.get_group_id(alignment, None) == "ACTGACTG"

    def test_get_group_missing_tag(self):
        """Test getting group when read lacks tag."""
        grouper = AlignmentTagReadGrouper("CB")
        alignment = MockAlignment()
        result = grouper.get_group_id(alignment, None)
        # Should return None
        assert result is None


class TestReadIdSplitReadGrouper:
    """Test ReadIdSplitReadGrouper class."""

    def test_init(self):
        """Test ReadIdSplitReadGrouper initialization."""
        grouper = ReadIdSplitReadGrouper("_")
        assert grouper.delim == "_"

    def test_get_group_with_delimiter(self):
        """Test getting group from read ID with delimiter."""
        grouper = ReadIdSplitReadGrouper("_")

        alignment1 = MockAlignment(read_id="read_001_groupA")
        alignment2 = MockAlignment(read_id="read_002_groupB")

        assert grouper.get_group_id(alignment1) == "groupA"
        assert grouper.get_group_id(alignment2) == "groupB"

    def test_get_group_no_delimiter(self):
        """Test getting group from read ID without delimiter."""
        grouper = ReadIdSplitReadGrouper("_")
        alignment = MockAlignment(read_id="read001")

        # Returns empty string if no delimiter found
        result = grouper.get_group_id(alignment)
        assert result == ""

    def test_get_group_multiple_delimiters(self):
        """Test getting group with multiple delimiters."""
        grouper = ReadIdSplitReadGrouper("_")
        alignment = MockAlignment(read_id="prefix_middle_suffix")

        # Should return last part after delimiter
        assert grouper.get_group_id(alignment) == "suffix"


class TestReadTableGrouper:
    """Test ReadTableGrouper class."""

    def test_init_and_get_group(self):
        """Test ReadTableGrouper initialization and file loading."""
        # Create temporary file with read-group mapping
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("read_001\tgroupA\n")
            f.write("read_002\tgroupB\n")
            f.write("read_003\tgroupA\n")
            temp_file = f.name

        try:
            from src.read_groups import SharedTableData, ReadTableGrouper

            # Create shared data with single column
            shared_data = SharedTableData(temp_file,
                                         read_id_column_index=0,
                                         group_id_column_indices=[1],
                                         delim='\t')
            grouper = ReadTableGrouper(shared_data, 0)

            alignment1 = MockAlignment(read_id="read_001")
            alignment2 = MockAlignment(read_id="read_002")
            alignment3 = MockAlignment(read_id="read_003")

            assert grouper.get_group_id(alignment1) == "groupA"
            assert grouper.get_group_id(alignment2) == "groupB"
            assert grouper.get_group_id(alignment3) == "groupA"
        finally:
            os.unlink(temp_file)

    def test_missing_read(self):
        """Test getting group for read not in file."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("read_001\tgroupA\n")
            temp_file = f.name

        try:
            from src.read_groups import SharedTableData, ReadTableGrouper

            # Create shared data with single column
            shared_data = SharedTableData(temp_file,
                                         read_id_column_index=0,
                                         group_id_column_indices=[1],
                                         delim='\t')
            grouper = ReadTableGrouper(shared_data, 0)

            alignment = MockAlignment(read_id="read_999")
            # Should return "NA" for missing reads
            result = grouper.get_group_id(alignment)
            assert result == "NA"
        finally:
            os.unlink(temp_file)


class TestSharedTableData:
    """Test SharedTableData and ReadTableGrouper classes."""

    def test_shared_data_multicolumn(self):
        """Test shared table data with multiple columns."""
        # Create temporary file with multi-column read-group mapping
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("read_001\tcell_A\tspot_1\ttissue_X\n")
            f.write("read_002\tcell_B\tspot_2\ttissue_Y\n")
            f.write("read_003\tcell_A\tspot_1\ttissue_X\n")
            temp_file = f.name

        try:
            from src.read_groups import SharedTableData, ReadTableGrouper

            # Create shared data with 3 group columns (indices 1, 2, 3)
            shared_data = SharedTableData(temp_file,
                                         read_id_column_index=0,
                                         group_id_column_indices=[1, 2, 3],
                                         delim='\t')

            # Create separate groupers for each column
            grouper_col1 = ReadTableGrouper(shared_data, 0)  # cell type
            grouper_col2 = ReadTableGrouper(shared_data, 1)  # spot
            grouper_col3 = ReadTableGrouper(shared_data, 2)  # tissue

            alignment1 = MockAlignment(read_id="read_001")
            alignment2 = MockAlignment(read_id="read_002")
            alignment3 = MockAlignment(read_id="read_003")

            # Test column 1 (cell type)
            assert grouper_col1.get_group_id(alignment1) == "cell_A"
            assert grouper_col1.get_group_id(alignment2) == "cell_B"
            assert grouper_col1.get_group_id(alignment3) == "cell_A"

            # Test column 2 (spot)
            assert grouper_col2.get_group_id(alignment1) == "spot_1"
            assert grouper_col2.get_group_id(alignment2) == "spot_2"
            assert grouper_col2.get_group_id(alignment3) == "spot_1"

            # Test column 3 (tissue)
            assert grouper_col3.get_group_id(alignment1) == "tissue_X"
            assert grouper_col3.get_group_id(alignment2) == "tissue_Y"
            assert grouper_col3.get_group_id(alignment3) == "tissue_X"

            # Verify all three groupers share the same data object
            assert grouper_col1.shared_data is shared_data
            assert grouper_col2.shared_data is shared_data
            assert grouper_col3.shared_data is shared_data

        finally:
            os.unlink(temp_file)

    def test_shared_data_missing_read(self):
        """Test shared data with missing read."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("read_001\tgroupA\tgroupX\n")
            temp_file = f.name

        try:
            from src.read_groups import SharedTableData, ReadTableGrouper

            shared_data = SharedTableData(temp_file,
                                         read_id_column_index=0,
                                         group_id_column_indices=[1, 2],
                                         delim='\t')

            grouper = ReadTableGrouper(shared_data, 0)
            alignment = MockAlignment(read_id="read_999")

            # Should return "NA" for missing reads
            result = grouper.get_group_id(alignment)
            assert result == "NA"

        finally:
            os.unlink(temp_file)


class TestParseGroupingSpec:
    """Test parse_grouping_spec function."""

    def test_parse_multicolumn_creates_separate_groupers(self):
        """Test that comma-separated columns create separate groupers sharing data."""
        # Create temporary file with multi-column data
        # Note: parse_grouping_spec appends "_spec{index}_chr_id" to the file path
        import tempfile
        temp_dir = tempfile.mkdtemp()
        base_file = os.path.join(temp_dir, "test_groups")
        chr_id = "chr1"
        spec_index = 0
        chr_file = base_file + "_spec" + str(spec_index) + "_" + chr_id

        with open(chr_file, 'w') as f:
            f.write("read_001\tcell_A\tspot_1\n")
            f.write("read_002\tcell_B\tspot_2\n")

        try:
            from src.read_groups import parse_grouping_spec, ReadTableGrouper

            # Mock args and sample objects
            class MockArgs:
                pass

            class MockSample:
                def __init__(self, file_path):
                    self.read_group_file = file_path

            args = MockArgs()
            sample = MockSample(base_file)  # Use base file, function will append spec_index and chr_id

            # Parse specification with comma-separated columns: file:path:0:1,2
            # Note: the file path in spec_string is ignored, sample.read_group_file is used
            spec_string = f"file:dummy:0:1,2"
            result = parse_grouping_spec(spec_string, args, sample, chr_id, spec_index)

            # Should return a list of two groupers
            assert isinstance(result, list)
            assert len(result) == 2

            # Both should be ReadTableGrouper instances
            assert isinstance(result[0], ReadTableGrouper)
            assert isinstance(result[1], ReadTableGrouper)

            # Both should share the same SharedTableData
            assert result[0].shared_data is result[1].shared_data

            # Column indices should be 0 and 1
            assert result[0].column_index == 0
            assert result[1].column_index == 1

            # Verify they return correct values
            alignment1 = MockAlignment(read_id="read_001")
            alignment2 = MockAlignment(read_id="read_002")

            assert result[0].get_group_id(alignment1) == "cell_A"
            assert result[0].get_group_id(alignment2) == "cell_B"
            assert result[1].get_group_id(alignment1) == "spot_1"
            assert result[1].get_group_id(alignment2) == "spot_2"

        finally:
            import shutil
            shutil.rmtree(temp_dir)

    def test_multiple_different_files(self):
        """Test that multiple different table files use separate split files."""
        import tempfile
        temp_dir = tempfile.mkdtemp()
        base_file = os.path.join(temp_dir, "test_groups")
        chr_id = "chr1"

        # Create split files for two different specs
        spec0_file = base_file + "_spec0_" + chr_id
        spec1_file = base_file + "_spec1_" + chr_id

        with open(spec0_file, 'w') as f:
            f.write("read_001\tgroupA\n")
            f.write("read_002\tgroupB\n")

        with open(spec1_file, 'w') as f:
            f.write("read_001\tgroupX\n")
            f.write("read_002\tgroupY\n")

        try:
            from src.read_groups import parse_grouping_spec, ReadTableGrouper

            class MockArgs:
                pass

            class MockSample:
                def __init__(self, file_path):
                    self.read_group_file = file_path

            args = MockArgs()
            sample = MockSample(base_file)

            # Parse two different file specs
            spec0_string = "file:groups1.tsv:0:1"
            spec1_string = "file:groups2.tsv:0:1"

            grouper0 = parse_grouping_spec(spec0_string, args, sample, chr_id, spec_index=0)
            grouper1 = parse_grouping_spec(spec1_string, args, sample, chr_id, spec_index=1)

            # Both should be ReadTableGrouper instances
            assert isinstance(grouper0, ReadTableGrouper)
            assert isinstance(grouper1, ReadTableGrouper)

            # They should use different underlying data
            assert grouper0.shared_data is not grouper1.shared_data

            # Verify they return different values for the same read
            alignment1 = MockAlignment(read_id="read_001")
            alignment2 = MockAlignment(read_id="read_002")

            assert grouper0.get_group_id(alignment1) == "groupA"
            assert grouper0.get_group_id(alignment2) == "groupB"

            assert grouper1.get_group_id(alignment1) == "groupX"
            assert grouper1.get_group_id(alignment2) == "groupY"

        finally:
            import shutil
            shutil.rmtree(temp_dir)


class TestParseBarcodeSpotSpec:
    """Test parse_barcode2spot_spec function."""

    def test_default_columns(self):
        """Test parsing spec with default columns."""
        from src.read_groups import parse_barcode2spot_spec

        filename, barcode_col, spot_cols = parse_barcode2spot_spec("my_file.tsv")

        assert filename == "my_file.tsv"
        assert barcode_col == 0
        assert spot_cols == [1]

    def test_explicit_single_column(self):
        """Test parsing spec with explicit single column."""
        from src.read_groups import parse_barcode2spot_spec

        filename, barcode_col, spot_cols = parse_barcode2spot_spec("my_file.tsv:0:1")

        assert filename == "my_file.tsv"
        assert barcode_col == 0
        assert spot_cols == [1]

    def test_explicit_multiple_columns(self):
        """Test parsing spec with explicit multiple columns."""
        from src.read_groups import parse_barcode2spot_spec

        filename, barcode_col, spot_cols = parse_barcode2spot_spec("my_file.tsv:0:1,2,3")

        assert filename == "my_file.tsv"
        assert barcode_col == 0
        assert spot_cols == [1, 2, 3]

    def test_custom_barcode_column(self):
        """Test parsing spec with custom barcode column."""
        from src.read_groups import parse_barcode2spot_spec

        filename, barcode_col, spot_cols = parse_barcode2spot_spec("my_file.tsv:2:3,4")

        assert filename == "my_file.tsv"
        assert barcode_col == 2
        assert spot_cols == [3, 4]


class TestBarcodeSpotGrouper:
    """Test BarcodeSpotGrouper class."""

    # Mock read_assignment object
    class MockReadAssignment:
        def __init__(self, barcode=None):
            self.barcode = barcode

    def test_init_and_get_group(self):
        """Test BarcodeSpotGrouper initialization and grouping."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("ACTGACTG\tcell_typeA\n")
            f.write("TGCATGCA\tcell_typeB\n")
            f.write("GGGGGGGG\tcell_typeA\n")
            temp_file = f.name

        try:
            from src.read_groups import SharedTableData, BarcodeSpotGrouper

            # Create shared data (barcode col 0, spot col 1)
            shared_data = SharedTableData(temp_file,
                                         read_id_column_index=0,
                                         group_id_column_indices=[1],
                                         delim='\t')
            grouper = BarcodeSpotGrouper(shared_data, column_index=0)

            # Create read assignments with barcodes
            ra1 = self.MockReadAssignment(barcode="ACTGACTG")
            ra2 = self.MockReadAssignment(barcode="TGCATGCA")
            ra3 = self.MockReadAssignment(barcode="GGGGGGGG")

            # Alignment is ignored, barcode comes from read_assignment
            alignment = MockAlignment()

            assert grouper.get_group_id(alignment, read_assignment=ra1) == "cell_typeA"
            assert grouper.get_group_id(alignment, read_assignment=ra2) == "cell_typeB"
            assert grouper.get_group_id(alignment, read_assignment=ra3) == "cell_typeA"

        finally:
            os.unlink(temp_file)

    def test_missing_barcode(self):
        """Test getting group when barcode is not in mapping."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("ACTGACTG\tcell_typeA\n")
            temp_file = f.name

        try:
            from src.read_groups import SharedTableData, BarcodeSpotGrouper

            shared_data = SharedTableData(temp_file,
                                         read_id_column_index=0,
                                         group_id_column_indices=[1],
                                         delim='\t')
            grouper = BarcodeSpotGrouper(shared_data, column_index=0)

            # Unknown barcode
            ra = self.MockReadAssignment(barcode="ZZZZZZZZ")
            alignment = MockAlignment()

            result = grouper.get_group_id(alignment, read_assignment=ra)
            assert result == "NA"

        finally:
            os.unlink(temp_file)

    def test_no_barcode(self):
        """Test getting group when read_assignment has no barcode."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("ACTGACTG\tcell_typeA\n")
            temp_file = f.name

        try:
            from src.read_groups import SharedTableData, BarcodeSpotGrouper

            shared_data = SharedTableData(temp_file,
                                         read_id_column_index=0,
                                         group_id_column_indices=[1],
                                         delim='\t')
            grouper = BarcodeSpotGrouper(shared_data, column_index=0)

            # No barcode
            ra = self.MockReadAssignment(barcode=None)
            alignment = MockAlignment()

            result = grouper.get_group_id(alignment, read_assignment=ra)
            assert result == "NA"

        finally:
            os.unlink(temp_file)

    def test_multicolumn_barcode_spot(self):
        """Test BarcodeSpotGrouper with multiple columns."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.tsv') as f:
            f.write("ACTGACTG\torgan_liver\tcell_hepatocyte\tzone_central\n")
            f.write("TGCATGCA\torgan_kidney\tcell_nephron\tzone_cortex\n")
            temp_file = f.name

        try:
            from src.read_groups import SharedTableData, BarcodeSpotGrouper

            # Create shared data with 3 spot columns
            shared_data = SharedTableData(temp_file,
                                         read_id_column_index=0,
                                         group_id_column_indices=[1, 2, 3],
                                         delim='\t')

            # Create groupers for each column
            grouper_col0 = BarcodeSpotGrouper(shared_data, column_index=0)
            grouper_col1 = BarcodeSpotGrouper(shared_data, column_index=1)
            grouper_col2 = BarcodeSpotGrouper(shared_data, column_index=2)

            ra1 = self.MockReadAssignment(barcode="ACTGACTG")
            ra2 = self.MockReadAssignment(barcode="TGCATGCA")
            alignment = MockAlignment()

            # Test column 0 (organ)
            assert grouper_col0.get_group_id(alignment, read_assignment=ra1) == "organ_liver"
            assert grouper_col0.get_group_id(alignment, read_assignment=ra2) == "organ_kidney"

            # Test column 1 (cell)
            assert grouper_col1.get_group_id(alignment, read_assignment=ra1) == "cell_hepatocyte"
            assert grouper_col1.get_group_id(alignment, read_assignment=ra2) == "cell_nephron"

            # Test column 2 (zone)
            assert grouper_col2.get_group_id(alignment, read_assignment=ra1) == "zone_central"
            assert grouper_col2.get_group_id(alignment, read_assignment=ra2) == "zone_cortex"

            # Verify all groupers share the same data
            assert grouper_col0.shared_data is shared_data
            assert grouper_col1.shared_data is shared_data
            assert grouper_col2.shared_data is shared_data

        finally:
            os.unlink(temp_file)


if __name__ == '__main__':
    pytest.main([__file__, '-v'])




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import os
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np
import pprint


class PlotOutput:
    def __init__(
        self,
        updated_gene_dict,
        gene_names,
        output_directory,
        create_visualization_subdir=False,
        reads_and_class=None,
        filter_transcripts=None,
        conditions=False,
        use_counts=False,
    ):
        self.updated_gene_dict = updated_gene_dict
        self.gene_names = gene_names
        self.output_directory = output_directory
        self.reads_and_class = reads_and_class
        self.filter_transcripts = filter_transcripts
        self.conditions = conditions
        self.use_counts = use_counts

        # Create visualization subdirectory if specified
        if create_visualization_subdir:
            self.visualization_dir = os.path.join(
                self.output_directory, "visualization"
            )
            os.makedirs(self.visualization_dir, exist_ok=True)
        else:
            self.visualization_dir = self.output_directory

    def plot_transcript_map(self):
        # Get the first condition's gene dictionary
        first_condition = next(iter(self.updated_gene_dict))
        gene_dict = self.updated_gene_dict[first_condition]

        for gene_name in self.gene_names:
            if gene_name in gene_dict:
                gene_data = gene_dict[gene_name]
                num_transcripts = len(gene_data["transcripts"])
                plot_height = max(
                    3, num_transcripts * 0.3
                )  # Adjust the height dynamically

                fig, ax = plt.subplots(
                    figsize=(12, plot_height)
                )  # Adjust height dynamically

                if self.filter_transcripts is not None:
                    ax.set_title(
                        f"Transcripts of Gene: {gene_data['name']} on Chromosome {gene_data['chromosome']} with value over {self.filter_transcripts}"
                    )
                else:
                    ax.set_title(
                        f"Transcripts of Gene: {gene_data['name']} on Chromosome {gene_data['chromosome']}"
                    )

                ax.set_xlabel("Chromosomal position")
                ax.set_ylabel("Transcripts")
                ax.set_yticks(range(num_transcripts))
                ax.set_yticklabels(
                    [
                        f"{transcript_id}"
                        for transcript_id in gene_data["transcripts"].keys()
                    ]
                )

                ax.xaxis.set_major_locator(
                    ticker.MaxNLocator(integer=True)
                )  # Ensure genomic positions are integers
                ax.xaxis.set_major_formatter(
                    ticker.FuncFormatter(lambda x, pos: f"{int(x)}")
                )  # Format x-axis ticks as integers

                # Plot each transcript
                for i, (transcript_id, transcript_info) in enumerate(
                    gene_data["transcripts"].items()
                ):
                    # Determine the direction based on the gene's strand information
                    direction_marker = ">" if gene_data["strand"] == "+" else "<"
                    marker_pos = (
                        transcript_info["end"] + 100
                        if gene_data["strand"] == "+"
                        else transcript_info["start"] - 100
                    )
                    ax.plot(
                        marker_pos,
                        i,
                        marker=direction_marker,
                        markersize=5,
                        color="blue",
                    )

                    # Draw the line for the whole transcript
                    ax.plot(
                        [transcript_info["start"], transcript_info["end"]],
                        [i, i],
                        color="grey",
                        linewidth=2,
                    )

                    # Exon blocks
                    for exon in transcript_info["exons"]:
                        exon_length = exon["end"] - exon["start"]
                        ax.add_patch(
                            plt.Rectangle(
                                (exon["start"], i - 0.4),
                                exon_length,
                                0.8,
                                color="skyblue",
                            )
                        )

                ax.set_xlim(gene_data["start"], gene_data["end"])
                ax.invert_yaxis()  # First transcript at the top

                plt.tight_layout()
                plot_path = os.path.join(
                    self.visualization_dir, f"{gene_name}_splicing.png"
                )
                plt.savefig(plot_path)  # Saving plot by gene name
                plt.close(fig)

    def plot_transcript_usage(self):
        """
        Visualize transcript usage for each gene in gene_names across different conditions.
        """

        for gene_name in self.gene_names:
            gene_data = {}
            for condition, genes in self.updated_gene_dict.items():
                if gene_name in genes:
                    gene_data[condition] = genes[gene_name]["transcripts"]

            if not gene_data:
                print(f"Gene {gene_name} not found in the data.")
                continue

            conditions = list(gene_data.keys())
            n_bars = len(conditions)

            fig, ax = plt.subplots(figsize=(12, 8))
            index = np.arange(n_bars)
            bar_width = 0.35
            opacity = 0.8

            # for sample_type, transcripts in gene_data.items():
            # print(f"Sample Type: {sample_type}")
            # for transcript_id, transcript_info in transcripts.items():
            # print(
            # f"  Transcript ID: {transcript_id}, Value: {transcript_info['value']}"
            # )
            # Adjusting the colors for better within-bar comparison
            max_transcripts = max(len(gene_data[condition]) for condition in conditions)
            colors = plt.cm.plasma(
                np.linspace(0, 1, num=max_transcripts)
            )  # Using plasma for better color gradation

            bottom_val = np.zeros(n_bars)
            for i, condition in enumerate(conditions):
                transcripts = gene_data[condition]
                for j, (transcript_id, transcript_info) in enumerate(
                    transcripts.items()
                ):
                    color = colors[j % len(colors)]
                    value = transcript_info["value"]
                    plt.bar(
                        i,
                        float(value),
                        bar_width,
                        bottom=bottom_val[i],
                        alpha=opacity,
                        color=color,
                        label=transcript_id if i == 0 else "",
                    )
                    bottom_val[i] += float(value)

            plt.xlabel("Sample Type")
            plt.ylabel("Transcript Usage (TPM)")
            plt.title(f"Transcript Usage for {gene_name} by Sample Type")
            plt.xticks(index, conditions)
            plt.legend(
                title="Transcript IDs", bbox_to_anchor=(1.05, 1), loc="upper left"
            )

            plt.tight_layout()
            plot_path = os.path.join(
                self.visualization_dir,
                f"{gene_name}_transcript_usage_by_sample_type.png",
            )
            plt.savefig(plot_path)
            plt.close(fig)

    def make_pie_charts(self):
        """
        Create pie charts for transcript alignment classifications and read assignment consistency.
        Handles both combined and separate sample data structures.
        """
        print("self.reads_and_class structure:")
        pprint.pprint(self.reads_and_class)

        titles = ["Transcript Alignment Classifications", "Read Assignment Consistency"]

        for title, data in zip(titles, self.reads_and_class):
            if isinstance(data, dict):
                if any(isinstance(v, dict) for v in data.values()):
                    # Separate 'Mutants' and 'WildType' case
                    for sample_name, sample_data in data.items():
                        self._create_pie_chart(f"{title} - {sample_name}", sample_data)
                else:
                    # Combined data case
                    self._create_pie_chart(title, data)
            else:
                print(f"Skipping unexpected data type for {title}: {type(data)}")

    def _create_pie_chart(self, title, data):
        """
        Helper method to create a single pie chart.
        """
        labels = list(data.keys())
        sizes = list(data.values())
        total = sum(sizes)

        # Generate a file-friendly title
        file_title = title.lower().replace(" ", "_").replace("-", "_")

        plt.figure(figsize=(12, 8))
        wedges, texts, autotexts = plt.pie(
            sizes,
            labels=labels,
            autopct=lambda pct: f"{pct:.1f}%\n({int(pct/100.*total):d})",
            startangle=140,
            textprops=dict(color="w"),
        )
        plt.setp(autotexts, size=8, weight="bold")
        plt.setp(texts, size=7)

        plt.axis("equal")  # Equal aspect ratio ensures that pie is drawn as a circle.
        plt.title(f"{title}\nTotal: {total}")

        plt.legend(
            wedges,
            labels,
            title="Categories",
            loc="center left",
            bbox_to_anchor=(1, 0, 0.5, 1),
            fontsize=8,
        )
        plot_path = os.path.join(self.visualization_dir, f"{file_title}_pie_chart.png")
        plt.savefig(plot_path, bbox_inches="tight", dpi=300)
        plt.close()



############################################################################
# Copyright (c) 2024-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import math
from collections import defaultdict
from queue import PriorityQueue

from .common import interval_len, junctions_from_blocks, overlaps
from .gene_info import GeneInfo, StrandDetector
from .long_read_assigner import LongReadAssigner
from .long_read_profiles import CombinedProfileConstructor
from .isoform_assignment import (
    IsoformMatch,
    MatchClassification,
    MatchEvent,
    MatchEventSubtype,
    ReadAssignment,
    ReadAssignmentType,
)
from .read_groups import DefaultReadGrouper
from .polya_finder import PolyAFinder
from .polya_verification import PolyAFixer
from .exon_corrector import ExonCorrector
from .alignment_info import AlignmentInfo
from .illumina_exon_corrector import IlluminaExonCorrector, VoidExonCorrector
from .stats import EnumStats
from enum import Enum, unique

logger = logging.getLogger('IsoQuant')


@unique
class PolyATrimmed(Enum):
    none = 1
    all = 2
    stranded = 3
    # file = 3


@unique
class AlignmentType(Enum):
    primary = 1
    secondary = 2
    supplementary = 20
    unaligned = 255


def make_alignment_tuple(bam_index, alignment):
    return alignment.reference_start, alignment.reference_end, bam_index, alignment


class BAMOnlineMerger:
    # single interator for several bam files
    def __init__(self, bam_pairs, chr_id, start, end, multiple_iterators=False):
        self.bam_pairs = bam_pairs
        self.multiple_iterators = multiple_iterators
        self._set(chr_id, start, end)

    def _set(self, chr_id, start, end):
        self.chr_id = chr_id
        self.start = start
        self.end = end
        # fetch uses 0-based semi-closed interval
        self.alignment_iterators = [bp[0].fetch(self.chr_id, self.start, self.end + 1,
                                                multiple_iterators=self.multiple_iterators) for bp in self.bam_pairs]
        self.current_elements = PriorityQueue(len(self.alignment_iterators))
        for i, it in enumerate(self.alignment_iterators):
            try:
                self.current_elements.put_nowait(make_alignment_tuple(i, next(it)))
            except StopIteration:
                pass

    def get(self):
        while not self.current_elements.empty():
            alignment_tuple = self.current_elements.get_nowait()
            i = alignment_tuple[2]
            try:
                self.current_elements.put_nowait(make_alignment_tuple(i, next(self.alignment_iterators[i])))
            except StopIteration:
                pass
            yield i, alignment_tuple[3]

    def reset(self):
        for bp in self.bam_pairs:
            bp[0].reset()

    def reset_region(self, chr_id, start, end):
        self._set(chr_id, start, end)


class AbstractAlignmentStorage:
    COVERAGE_BIN = 256

    def __init__(self):
        self.coverage_dict = defaultdict(int)
        self.region = None

    def reset(self):
        self.coverage_dict = defaultdict(int)
        self.region = None

    def add_alignment(self, bam_index, alignment):
        bin_start = alignment.reference_start // AbstractAlignmentStorage.COVERAGE_BIN
        # closed interval, 0-based
        alignment_end = alignment.reference_end - 1
        bin_end = alignment_end // AbstractAlignmentStorage.COVERAGE_BIN
        for i in range(bin_start, bin_end + 1):
            self.coverage_dict[i] += 1
        if not self.region:
            self.region = (alignment.reference_start, alignment_end)
        else:
            self.region = (min(self.region[0], alignment.reference_start), max(self.region[1], alignment_end))

    def get_alignments(self, region=None):
        raise NotImplementedError()

    def get_read_count(self):
        raise NotImplementedError()

    def alignment_is_not_adjacent(self, alignment):
        return self.region is not None and \
            not overlaps(self.region, (alignment.reference_start, alignment.reference_end - 1))


class BAMAlignmentStorage(AbstractAlignmentStorage):
    def __init__(self, bam_merger):
        AbstractAlignmentStorage.__init__(self)
        self.bam_merger = bam_merger
        self.counter = 0

    def reset(self):
        AbstractAlignmentStorage.reset(self)
        self.counter = 0

    def add_alignment(self, bam_index, alignment):
        AbstractAlignmentStorage.add_alignment(self, bam_index, alignment)
        self.counter += 1

    def get_alignments(self, region=None):
        if not region:
            region = self.region
        return BAMOnlineMerger(self.bam_merger.bam_pairs, self.bam_merger.chr_id, region[0], region[1],
                               multiple_iterators=True).get()

    def get_read_count(self):
        return self.counter


class InMemoryAlignmentStorage(AbstractAlignmentStorage):
    def __init__(self):
        AbstractAlignmentStorage.__init__(self)
        self.alignment_start_index = {}
        self.alignment_end_index = {}
        self.counter = 0
        self.alignment_storage = []
        self.index_filled = False

    def reset(self):
        AbstractAlignmentStorage.reset(self)
        self.alignment_start_index = {}
        self.alignment_end_index = {}
        self.counter = 0
        self.alignment_storage = []
        self.index_filled = False

    def add_alignment(self, bam_index, alignment):
        AbstractAlignmentStorage.add_alignment(self, bam_index, alignment)
        bin_start_position = alignment.reference_start // self.COVERAGE_BIN
        if bin_start_position not in self.alignment_start_index:
            self.alignment_start_index[bin_start_position] = self.counter
        bin_end_position = (alignment.reference_end - 1) // self.COVERAGE_BIN
        if bin_end_position not in self.alignment_end_index:
            self.alignment_end_index[bin_end_position] = self.counter
        self.alignment_storage.append((bam_index, alignment))
        self.counter += 1
        self.index_filled = False

    def fill_index(self):
        if self.index_filled:
            return
        current_index = len(self.alignment_storage)
        current_bin_region_start = self.region[0] // AbstractAlignmentStorage.COVERAGE_BIN
        current_bin_region_end = self.region[1] // AbstractAlignmentStorage.COVERAGE_BIN
        for pos in range(current_bin_region_end + 1, current_bin_region_start - 1, -1):
            if pos not in self.alignment_start_index:
                self.alignment_start_index[pos] = current_index
            else:
                current_index = self.alignment_start_index[pos]
        current_index = len(self.alignment_storage)
        for pos in range(current_bin_region_end + 1, current_bin_region_start - 1, -1):
            if pos not in self.alignment_end_index or self.alignment_end_index[pos] > current_index:
                self.alignment_end_index[pos] = current_index
            else:
                current_index = self.alignment_end_index[pos]
        self.index_filled = True

    def get_alignments(self, region=None):
        if not region or region == self.region:
            for a in self.alignment_storage:
                yield a
            return

        self.fill_index()
        # first alignment among sorted that has its end inside the start_bin, e.g. close to region[0]
        start_bin = region[0] // self.COVERAGE_BIN
        start_index = self.alignment_end_index[start_bin]
        # first alignment that has its start after region[1]
        end_bin = region[1] // self.COVERAGE_BIN
        end_index = self.alignment_start_index[end_bin]

        for i in range(start_index, end_index):
            bam_index, alignment = self.alignment_storage[i]
            if overlaps(region, (alignment.reference_start, alignment.reference_end - 1)):
                yield bam_index, alignment

    def get_read_count(self):
        return len(self.alignment_storage)


class AlignmentCollector:
    """ class for aggregating all alignmnet information

    Parameters
    ----------
    bams : pair bam object, filename
    params
    printer
    counter
    """

    MIN_REGION_LEN = 32768
    MAX_REGION_LEN = 1048576
    MIN_READS_TO_SPLIT = 1024
    ABS_COV_VALLEY = 1
    REL_COV_VALLEY = 0.01

    SMALL_CHR_IDS = ['MT', 'chrM', 'chrMT']
    SMALL_CHR_LEN = 500000
    WARN_COVERAGE = 2000000

    def __init__(self, chr_id, bam_pairs, params, illumina_bam,
                 genedb=None, chr_record=None, read_groupper=DefaultReadGrouper(),
                 barcode_dict=None,
                 small_chr_max_coverage=1000000,
                 usual_gene_max_coverage=-1,
                 string_pools=None):
        self.chr_id = chr_id
        self.bam_pairs = bam_pairs
        self.params = params
        self.genedb = genedb
        self.chr_record = chr_record
        self.illumina_bam = illumina_bam
        self.string_pools = string_pools

        self.bam_merger = BAMOnlineMerger(self.bam_pairs, self.chr_id, 0,
                                          self.bam_pairs[0][0].get_reference_length(self.chr_id),
                                          multiple_iterators=not self.params.high_memory)
        self.strand_detector = StrandDetector(self.chr_record)
        self.read_groupper = read_groupper
        self.barcode_dict = barcode_dict # read_id -> (barcode, umi)
        self.barcode_tag = params.barcode_tag if getattr(params, 'barcoded_bam', False) else None
        self.umi_tag = params.umi_tag if getattr(params, 'barcoded_bam', False) else None
        self.strip_barcode_suffix = getattr(params, 'strip_barcode_suffix', False)
        self.small_chr_max_coverage = small_chr_max_coverage
        self.usual_gene_max_coverage = usual_gene_max_coverage
        self.polya_finder = PolyAFinder(self.params.polya_window, self.params.polya_fraction)
        self.polya_fixer = PolyAFixer(self.params)
        # self.cage_finder = CagePeakFinder(params.cage, params.cage_shift)
        self.alignment_stat_counter = EnumStats()

    def process(self):
        alignment_storage = BAMAlignmentStorage(self.bam_merger) if not self.params.high_memory else InMemoryAlignmentStorage()
        for bam_index, alignment in self.bam_merger.get():
            if alignment.is_secondary:
                self.alignment_stat_counter.add(AlignmentType.secondary)
            elif alignment.is_supplementary:
                self.alignment_stat_counter.add(AlignmentType.supplementary)
            elif alignment.reference_id != -1:
                self.alignment_stat_counter.add(AlignmentType.primary)

            if alignment_storage.alignment_is_not_adjacent(alignment):
                preceding_genes = self.get_genes_in_region(alignment_storage.region)
                next_genes = self.get_genes_in_region((alignment.reference_start, alignment.reference_end - 1))

                if len(preceding_genes.intersection(next_genes)) == 0:
                    for res in self.forward_alignments(alignment_storage):
                        yield res
                    alignment_storage.reset()

            alignment_storage.add_alignment(bam_index, alignment)

        if alignment_storage.region:
            for res in self.forward_alignments(alignment_storage):
                yield res

    def forward_alignments(self, alignment_storage):
        current_region = alignment_storage.region
        logger.debug("Splitting " + str(current_region))
        split_regions = AlignmentCollector.split_coverage_regions(current_region, alignment_storage)

        if len(split_regions) == 1:
            max_coverage = split_regions[0][1]
            yield self.process_alignments_in_region(current_region, alignment_storage.get_alignments(), max_coverage)
        else:
            for new_region, max_coverage in split_regions:
                alignments = alignment_storage.get_alignments(new_region)
                yield self.process_alignments_in_region(new_region, alignments, max_coverage)

    def process_alignments_in_region(self, current_region, alignment_storage, max_coverage):
        logger.debug("Processing region %s" % str(current_region))

        skip_read_fraction = 1
        chr_len = len(self.chr_record)
        if self.chr_id in self.SMALL_CHR_IDS or len(self.chr_record) < self.SMALL_CHR_LEN:
            coverage_cutoff = self.small_chr_max_coverage
            chromosome_description = "small "
            option_string = "max_coverage_small_chr"
        else:
            coverage_cutoff = self.usual_gene_max_coverage
            chromosome_description = ""
            option_string = "max_coverage_normal_chr"

        if max_coverage > coverage_cutoff > 0:
            skip_read_fraction = math.ceil(max_coverage / coverage_cutoff)
            logger.warning("Genomic region %d-%d on %schromosome %s (region length %d, chromosome length %d) "
                           "has coverage %d, which exceed coverage cutoff %d" %
                           (current_region[0], current_region[1], chromosome_description, self.chr_id,
                            current_region[1] - current_region[0], chr_len, max_coverage, coverage_cutoff))
            logger.warning("Large number of reads mapped to a single loci may significantly "
                           "increase running time and RAM consumption")
            logger.warning("IsoQuant will process only 1 read out of every %d, "
                           "use --%s to change the coverage limit for small chromosomes" % (skip_read_fraction, option_string))
        elif max_coverage > self.WARN_COVERAGE:
            logger.info("Genomic region %d-%d on %schromosome %s (region length %d) has high coverage %d" %
                        (current_region[0], current_region[1], chromosome_description, self.chr_id,
                         current_region[1] - current_region[0], max_coverage))
            logger.info("Large number of reads mapped to a single loci may significantly "
                        "increase running time and RAM consumption, maximum coverage threshold "
                        "can be set via --%s" % option_string)

        gene_info = self.get_gene_info_for_region(current_region)
        if gene_info.empty():
            assignment_storage = self.process_intergenic(alignment_storage, current_region, skip_read_fraction)
        else:
            assignment_storage = self.process_genic(alignment_storage, gene_info, current_region, skip_read_fraction)

        return gene_info, assignment_storage

    def process_intergenic(self, alignment_storage, region, skip_read_fraction=1):
        if self.illumina_bam is not None:
            corrector = IlluminaExonCorrector(self.chr_id, region[0], region[1], self.illumina_bam)
        else:
            corrector = VoidExonCorrector()

        counter = 0
        for bam_index, alignment in alignment_storage:
            if alignment.reference_id == -1 or alignment.is_supplementary or \
                    (not self.params.use_secondary and alignment.is_secondary):
                continue

            if self.params.min_mapq and alignment.mapping_quality < self.params.min_mapq:
                continue

            read_id = alignment.query_name
            alignment_info = AlignmentInfo(alignment)

            if not alignment_info.read_exons:
                logger.warning("Read %s has no aligned exons" % read_id)
                continue

            if len(alignment_info.read_exons) <= 2 and \
                    (alignment.is_secondary or alignment.mapping_quality < self.params.simple_alignments_mapq_cutoff):
                continue

            counter += 1
            if skip_read_fraction > 1 and counter % skip_read_fraction != 0:
                continue

            if self.params.polya_trimmed == PolyATrimmed.none:
                alignment_info.add_polya_info(self.polya_finder, self.polya_fixer)

            read_assignment = ReadAssignment(read_id, ReadAssignmentType.intergenic, self.string_pools,
                                             match=IsoformMatch(MatchClassification.intergenic, string_pools=self.string_pools))

            if alignment_info.exons_changed:
                read_assignment.add_match_attribute(MatchEvent(MatchEventSubtype.aligned_polya_tail))

            read_assignment.polya_info = alignment_info.polya_info
            read_assignment.cage_found = len(alignment_info.cage_hits) > 0
            read_assignment.genomic_region = region
            read_assignment.exons = alignment_info.read_exons
            read_assignment.corrected_exons = corrector.correct_read(alignment_info)
            read_assignment.corrected_introns = junctions_from_blocks(read_assignment.corrected_exons)

            # Populate barcode and UMI first (needed by some groupers)
            if read_id in self.barcode_dict:
                read_assignment.barcode, read_assignment.umi = self.barcode_dict[read_id]
            elif self.barcode_tag and alignment.has_tag(self.barcode_tag):
                barcode = alignment.get_tag(self.barcode_tag)
                if self.strip_barcode_suffix and '-' in barcode:
                    barcode = barcode.rsplit('-', 1)[0]
                read_assignment.barcode = barcode
                read_assignment.umi = alignment.get_tag(self.umi_tag) if alignment.has_tag(self.umi_tag) else None
            # Get group ID(s) - pass read_assignment for groupers that need barcode
            group_ids = self.read_groupper.get_group_id(alignment, read_assignment, self.bam_merger.bam_pairs[bam_index][1])
            # Ensure read_group is always a list
            read_assignment.read_group = group_ids if isinstance(group_ids, list) else [group_ids]
            read_assignment.mapped_strand = "-" if alignment.is_reverse else "+"
            read_assignment.strand = self.get_assignment_strand(read_assignment)
            read_assignment.chr_id = self.chr_id
            read_assignment.multimapper = alignment.is_secondary
            read_assignment.mapping_quality = alignment.mapping_quality

            self.add_artificial_polya(read_assignment)
            read_assignment.polyA_found = (alignment_info.polya_info.external_polya_pos != -1 or
                                           alignment_info.polya_info.external_polyt_pos != -1 or
                                           alignment_info.polya_info.internal_polya_pos != -1 or
                                           alignment_info.polya_info.internal_polyt_pos != -1)

            AlignmentCollector.import_bam_tags(alignment, read_assignment, self.params.bam_tags)
            yield read_assignment

    def process_genic(self, alignment_storage, gene_info, region, skip_read_fraction=1):
        assigner = LongReadAssigner(gene_info, self.params, self.string_pools)
        profile_constructor = CombinedProfileConstructor(gene_info, self.params)
        exon_corrector = ExonCorrector(gene_info, self.params, self.chr_record)

        counter = 0
        for bam_index, alignment in alignment_storage:
            if alignment.reference_id == -1 or alignment.is_supplementary or \
                    (not self.params.use_secondary and alignment.is_secondary):
                continue

            if self.params.min_mapq and alignment.mapping_quality < self.params.min_mapq:
                continue

            counter += 1
            if skip_read_fraction > 1 and counter % skip_read_fraction != 0:
                continue

            read_id = alignment.query_name
            logger.debug("=== Processing read " + read_id + " ===")
            alignment_info = AlignmentInfo(alignment)

            if not alignment_info.read_exons:
                logger.warning("Read %s has no aligned exons" % read_id)
                continue

            alignment_info.add_polya_info(self.polya_finder, self.polya_fixer)
            # if self.params.cage:
            #    alignment_info.add_cage_info(self.cage_finder)
            alignment_info.construct_profiles(profile_constructor)
            read_assignment = assigner.assign_to_isoform(read_id, alignment_info.combined_profile)

            if (not read_assignment.assignment_type in [ReadAssignmentType.unique,
                                                        ReadAssignmentType.unique_minor_difference,
                                                        ReadAssignmentType.ambiguous])\
                    and alignment.mapping_quality < self.params.inconsistent_mapq_cutoff:
                continue

            if alignment_info.exons_changed:
                read_assignment.add_match_attribute(MatchEvent(MatchEventSubtype.aligned_polya_tail))

            read_assignment.polya_info = alignment_info.polya_info
            read_assignment.cage_found = len(alignment_info.cage_hits) > 0
            read_assignment.genomic_region = region
            read_assignment.exons = alignment_info.read_exons
            read_assignment.corrected_exons = exon_corrector.correct_assigned_read(alignment_info,
                                                                                   read_assignment)
            read_assignment.corrected_introns = junctions_from_blocks(read_assignment.corrected_exons)

            # Populate barcode and UMI first (needed by some groupers)
            if read_id in self.barcode_dict:
                read_assignment.barcode, read_assignment.umi = self.barcode_dict[read_id]
            elif self.barcode_tag and alignment.has_tag(self.barcode_tag):
                barcode = alignment.get_tag(self.barcode_tag)
                if self.strip_barcode_suffix and '-' in barcode:
                    barcode = barcode.rsplit('-', 1)[0]
                read_assignment.barcode = barcode
                read_assignment.umi = alignment.get_tag(self.umi_tag) if alignment.has_tag(self.umi_tag) else None
            # Get group ID(s) - pass read_assignment for groupers that need barcode
            group_ids = self.read_groupper.get_group_id(alignment, read_assignment, self.bam_merger.bam_pairs[bam_index][1])
            # Ensure read_group is always a list
            read_assignment.read_group = group_ids if isinstance(group_ids, list) else [group_ids]
            read_assignment.mapped_strand = "-" if alignment.is_reverse else "+"
            read_assignment.strand = self.get_assignment_strand(read_assignment)
            AlignmentCollector.check_antisense(read_assignment)
            AlignmentCollector.import_bam_tags(alignment, read_assignment, self.params.bam_tags)
            read_assignment.chr_id = gene_info.chr_id
            read_assignment.multimapper = alignment.is_secondary
            read_assignment.mapping_quality = alignment.mapping_quality

            self.add_artificial_polya(read_assignment)
            read_assignment.polyA_found = (alignment_info.polya_info.external_polya_pos != -1 or
                                           alignment_info.polya_info.external_polyt_pos != -1 or
                                           alignment_info.polya_info.internal_polya_pos != -1 or
                                           alignment_info.polya_info.internal_polyt_pos != -1)

            if self.params.count_exons:
                read_assignment.exon_gene_profile = alignment_info.combined_profile.read_exon_profile.gene_profile
                read_assignment.intron_gene_profile = alignment_info.combined_profile.read_intron_profile.gene_profile

            logger.debug("=== Finished read " + read_id + " ===")
            yield read_assignment

    def get_assignment_strand(self, read_assignment):
        if read_assignment.isoform_matches and read_assignment.assignment_type in \
                [ReadAssignmentType.unique, ReadAssignmentType.unique_minor_difference]:
            return read_assignment.isoform_matches[0].transcript_strand

        has_polya = read_assignment.polya_info.external_polya_pos != -1 or read_assignment.polya_info.internal_polya_pos != -1
        has_polyt = read_assignment.polya_info.external_polyt_pos != -1 or read_assignment.polya_info.internal_polyt_pos != -1
        if len(read_assignment.exons) == 1:
            if has_polya and not has_polyt:
                return '+'
            elif has_polyt and not has_polya:
                return '-'
            return '.'
        return self.strand_detector.get_strand(read_assignment.corrected_introns, has_polya, has_polyt)

    def count_indel_stats(self, alignment):
        cigar_event_count = len(alignment.cigartuples)
        indel_events = [1, 2]
        indel_count = 0
        intron_cigar_positions = []
        for i in range(cigar_event_count):
            cigar = alignment.cigartuples[i]
            if cigar[0] in indel_events:
                indel_count += 1
            elif cigar[0] == 3:
                intron_cigar_positions.append(i)

        junctions_with_indels = 0
        for i in intron_cigar_positions:
            # indel right near intron
            if (i > 0 and alignment.cigartuples[i - 1][0] in indel_events) or \
                    (i < cigar_event_count - 1 and alignment.cigartuples[i + 1][0] in indel_events):
                junctions_with_indels += 1

            # indel separated by at most 'indel_near_splice_site_dist' matches from intron
            if (i > 1 and alignment.cigartuples[i - 2][0] in indel_events and
                alignment.cigartuples[i - 1][0] in [0, 7, 8] and
                alignment.cigartuples[i - 1][1] <= self.params.indel_near_splice_site_dist) or \
                    (i < cigar_event_count - 2 and alignment.cigartuples[i + 2][0] in indel_events and
                     alignment.cigartuples[i + 1][0]  in [0, 7, 8] and
                     alignment.cigartuples[i + 1][1] <= self.params.indel_near_splice_site_dist):
                junctions_with_indels += 1

        return indel_count, junctions_with_indels

    def get_genes_in_region(self, current_region):
        if not self.genedb:
            return set()
        return set(g.id for g in self.genedb.region(seqid=self.chr_id,
                                                    start=current_region[0],
                                                    end=current_region[1],
                                                    featuretype="gene"))

    def get_gene_info_for_region(self, current_region):
        if not self.genedb:
            return GeneInfo.from_region(self.chr_id, current_region[0], current_region[1],
                                        self.params.delta, self.chr_record)

        gene_list = list(self.genedb.region(seqid=self.chr_id, start=current_region[0],
                                            end=current_region[1], featuretype="gene"))
        if not gene_list:
            return GeneInfo.from_region(self.chr_id, current_region[0], current_region[1],
                                        self.params.delta, self.chr_record)
        gene_list = sorted(gene_list, key=lambda x: x.start)
        gene_info = GeneInfo(gene_list, self.genedb, self.params.delta)
        if self.params.needs_reference:
            gene_info.set_reference_sequence(current_region[0], current_region[1], self.chr_record)
        return gene_info

    def add_artificial_polya(self, read_assignment):
        if self.params.polya_trimmed == PolyATrimmed.stranded:
            if read_assignment.strand == '+':
                read_assignment.polya_info.external_polya_pos = read_assignment.corrected_exons[-1][1] + 1
            elif read_assignment.strand == '-':
                read_assignment.polya_info.external_polyt_pos = read_assignment.corrected_exons[0][0] - 1
        elif self.params.polya_trimmed == PolyATrimmed.all:
            if read_assignment.mapped_strand == '+':
                read_assignment.polya_info.external_polya_pos = read_assignment.corrected_exons[-1][1] + 1
            elif read_assignment.mapped_strand == '-':
                read_assignment.polya_info.external_polyt_pos = read_assignment.corrected_exons[0][0] - 1

    @staticmethod
    def split_coverage_regions(genomic_region, alignment_storage):
        if interval_len(genomic_region) < AlignmentCollector.MIN_REGION_LEN or \
                alignment_storage.get_read_count() < AlignmentCollector.MIN_READS_TO_SPLIT:
            max_coverage = max(alignment_storage.coverage_dict.values())
            return [(genomic_region, max_coverage)]

        split_regions = []
        coverage_dict = alignment_storage.coverage_dict
        coverage_positions = sorted(coverage_dict.keys())
        current_start = coverage_positions[0]
        min_bins = int(AlignmentCollector.MIN_REGION_LEN / AbstractAlignmentStorage.COVERAGE_BIN)
        max_bins = int(AlignmentCollector.MAX_REGION_LEN / AbstractAlignmentStorage.COVERAGE_BIN)
        pos = current_start + 1
        max_cov = coverage_dict[current_start]
        abs_cov_valley = AlignmentCollector.ABS_COV_VALLEY
        rel_cov_valley = AlignmentCollector.REL_COV_VALLEY
        prev_upd_pos = pos

        while pos <= coverage_positions[-1]:
            while (pos <= coverage_positions[-1] and
                   (pos - current_start < min_bins or
                    coverage_dict[pos] > max(abs_cov_valley,
                                             max_cov * rel_cov_valley))):
                if pos - prev_upd_pos > max_bins:
                    # increase valley cut-offs every MAX_REGION_LEN bases to avoid extra long regions
                    prev_upd_pos = pos
                    abs_cov_valley += AlignmentCollector.ABS_COV_VALLEY
                    rel_cov_valley += AlignmentCollector.REL_COV_VALLEY

                max_cov = max(max_cov, coverage_dict[pos])
                pos += 1

            split_regions.append(((max(current_start * AbstractAlignmentStorage.COVERAGE_BIN + 1, genomic_region[0]),
                                   min(pos * AbstractAlignmentStorage.COVERAGE_BIN, genomic_region[1])),
                                  max_cov))
            current_start = pos
            prev_upd_pos = pos
            abs_cov_valley = AlignmentCollector.ABS_COV_VALLEY
            rel_cov_valley = AlignmentCollector.REL_COV_VALLEY
            max_cov = coverage_dict[current_start]
            pos = min(current_start + 1, coverage_positions[-1] + 1)

        return split_regions

    @staticmethod
    def check_antisense(read_assignment):
        for match in read_assignment.isoform_matches:
            if match.assigned_transcript is None:
                continue

            if read_assignment.strand == match.transcript_strand or read_assignment.strand == '.':
                continue
            if match.match_classification not in \
                    [MatchClassification.novel_in_catalog, MatchClassification.novel_not_in_catalog]:
                match.match_classification = MatchClassification.antisense
            match.match_subclassifications.append(MatchEvent(MatchEventSubtype.antisense))

    @staticmethod
    def import_bam_tags(bam_record, read_assignment, tag_list):
        for tag in tag_list:
            try:
                tag_value = bam_record.get_tag(tag)
                read_assignment.set_additional_attribute(tag, str(tag_value))
            except KeyError:
                pass



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
from collections import defaultdict, OrderedDict
from enum import Enum, unique

from .isoform_assignment import (
    ReadAssignmentType,
)
from .gene_info import FeatureInfo
from .read_groups import AbstractReadGrouper
from .convert_grouped_counts import GROUP_COUNT_CUTOFF, convert_to_mtx, convert_to_matrix
from .file_naming import (
    counts_prefix, tpm_prefix, counts_file_name, tpm_file_name,
    counts_stats_file_name, counts_usable_file_name
)

logger = logging.getLogger('IsoQuant')


@unique
class CountingStrategy(Enum):
    unique_only = 1
    with_ambiguous = 2
    unique_splicing_consistent = 10
    unique_inconsistent = 11
    # all_splicing_consistent = 19
    all = 20

    def no_inconsistent(self):
        return self in [CountingStrategy.unique_only, CountingStrategy.with_ambiguous]

    def ambiguous(self):
        return self in [CountingStrategy.all, CountingStrategy.with_ambiguous]

    def inconsistent_minor(self):
        return self in [CountingStrategy.unique_splicing_consistent,
                        CountingStrategy.unique_inconsistent, CountingStrategy.all]

    def inconsistent(self):
        return self in [CountingStrategy.unique_inconsistent, CountingStrategy.all]


COUNTING_STRATEGIES = [CountingStrategy.unique_only.name,
                       CountingStrategy.with_ambiguous.name,
                       CountingStrategy.unique_splicing_consistent.name,
                       CountingStrategy.unique_inconsistent.name,
                       CountingStrategy.all.name]


class CountingStrategyFlags:
    def __init__(self, counting_strategy):
        self.use_ambiguous = counting_strategy.ambiguous()
        self.use_inconsistent_minor = counting_strategy.inconsistent_minor()
        self.use_inconsistent = counting_strategy.inconsistent()


@unique
class GroupedOutputFormat(Enum):
    none = 0
    matrix = 1
    mtx = 2
    default = 3


# the difference from defaultdict is that it does not add 0 to the dict when get is called for inexistent key
class IncrementalDict:
    def __init__(self, default_type=float):
        self.data = {}
        self.default_type = default_type

    def inc(self, key, value=1.0):
        if key not in self.data:
            self.data[key] = self.default_type(value)
        else:
            self.data[key] += value

    def get(self, key):
        if key not in self.data:
            return self.default_type()
        else:
            return self.data[key]


class GeneAssignmentExtractor:
    @staticmethod
    def get_features(read_assignment):
        gene_set = set()
        for m in read_assignment.isoform_matches:
            if m.assigned_gene:
                gene_set.add(m.assigned_gene)
        return gene_set

    @staticmethod
    def get_assignment_type(read_assignment):
        return read_assignment.gene_assignment_type

    @staticmethod
    def confirms_feature(read_assignment):
        return read_assignment.gene_assignment_type.is_unique()


class TranscriptAssignmentExtractor:
    @staticmethod
    def get_features(read_assignment):
        transcript_set = set()
        for m in read_assignment.isoform_matches:
            if m.assigned_transcript:
                transcript_set.add(m.assigned_transcript)
        return transcript_set

    @staticmethod
    def get_assignment_type(read_assignment):
        return read_assignment.assignment_type

    @staticmethod
    def confirms_feature(read_assignment):
        transcript_id = read_assignment.isoform_matches[0].assigned_transcript
        return (read_assignment.assignment_type.is_unique() and
                (len(read_assignment.gene_info.all_isoforms_introns[transcript_id]) == 0 or
                 len(read_assignment.corrected_exons) > 1))


class ReadWeightCounter:
    def __init__(self, strategy_str):
        self.strategy = CountingStrategy[strategy_str]
        self.strategy_flags = CountingStrategyFlags(self.strategy)

    def process_inconsistent(self, assignment_type, feature_count):
        # use only for inconsistent assignments
        if assignment_type == ReadAssignmentType.inconsistent_ambiguous or feature_count > 1:
            if self.strategy_flags.use_ambiguous and self.strategy_flags.use_inconsistent:
                return 1.0 / feature_count
            else:
                return 0.0
        if self.strategy_flags.use_inconsistent:
            return 1.0
        if self.strategy_flags.use_inconsistent_minor and assignment_type == ReadAssignmentType.inconsistent_non_intronic:
            return 1.0
        return 0.0

    def process_ambiguous(self, feature_count):
        if feature_count == 0:
            return 0.0
        if feature_count == 1:
            return 1.0
        if self.strategy_flags.use_ambiguous:
            return 1.0 / float(feature_count)
        else:
            return 0.0


class AbstractCounter:
    def __init__(self, output_prefix, ignore_read_groups=False):
        self.ignore_read_groups = ignore_read_groups
        self.output_prefix = output_prefix
        self.output_counts_prefix = counts_prefix(output_prefix)
        self.output_tpm_prefix = tpm_prefix(output_prefix)
        self.output_counts_file_name = counts_file_name(self.output_counts_prefix, linear=not ignore_read_groups)
        self.output_tpm_file_name = tpm_file_name(self.output_tpm_prefix)
        self.output_file = self.output_counts_file_name
        open(self.output_file, "w").close()
        self.output_stats_file_name = None
        self.usable_file_name = None

    def get_output_file_handler(self):
        return open(self.output_file, "a")

    def add_read_info(self, read_assignment):
        raise NotImplementedError()

    def add_read_info_raw(self, read_id, feature_ids, group_ids):
        raise NotImplementedError()

    def add_confirmed_features(self, features):
        raise NotImplementedError()

    def dump(self):
        raise NotImplementedError()

    def finalize(self, args=None):
        raise NotImplementedError()

    def add_unassigned(self, read_assignment):
        raise NotImplementedError()

    def add_unaligned(self, n_reads=1):
        raise NotImplementedError()


class CompositeCounter:
    def __init__(self, counters: list = None):
        if counters is None:
            self.counters = []
        else:
            self.counters: list = counters

    def add_counters(self, counters):
        self.counters += counters

    def add_counter(self, counter):
        self.counters.append(counter)

    def add_read_info(self, read_assignment):
        for p in self.counters:
            p.add_read_info(read_assignment)

    def add_read_info_raw(self, read_id, feature_ids, group_ids):
        for p in self.counters:
            p.add_read_info_raw(read_id, feature_ids, group_ids)

    def add_confirmed_features(self, features):
        for p in self.counters:
            p.add_confirmed_features(features)

    def dump(self):
        for p in self.counters:
            p.dump()

    def finalize(self, args=None):
        for p in self.counters:
            p.finalize(args)

    def add_unassigned(self, read_assignment):
        for p in self.counters:
            p.add_unassigned(read_assignment)

    def add_unaligned(self, n_reads=1):
        for p in self.counters:
            p.add_unaligned(n_reads)


# count meta-features assigned to reads (genes or isoforms)
# get_feature_id --- function that returns feature id form IsoformMatch object
class AssignedFeatureCounter(AbstractCounter):
    def __init__(self, output_prefix, assignment_extractor, string_pools, read_counter,
                 all_features=None, group_index: int = 0):
        AbstractCounter.__init__(self, output_prefix, string_pools is None)
        self.assignment_extractor = assignment_extractor
        self.string_pools = string_pools
        self.all_features = set(all_features) if all_features is not None else set()
        self.group_index = group_index  # Index in read_group list to use
        self.read_counter = read_counter

        self.ambiguous_reads = 0
        self.reads_for_tpm = defaultdict(int)
        self.not_assigned_reads = 0
        self.not_aligned_reads = 0
        # feature_id -> (group_id -> count)  -- group_id is now an integer pool index
        self.feature_counter = defaultdict(IncrementalDict)
        self.confirmed_features = set()
        self.output_stats_file_name = counts_stats_file_name(self.output_counts_file_name)
        self.usable_file_name = counts_usable_file_name(self.output_counts_file_name)

    def add_read_info(self, read_assignment=None):
        # Use read_group_ids directly (integers) instead of string conversion
        if self.ignore_read_groups:
            group_id = 0  # Default group index
        elif not read_assignment.read_group_ids:
            group_id = 0
        else:
            group_id = read_assignment.read_group_ids[self.group_index]

        if not read_assignment:
            self.not_aligned_reads += 1
            return
        elif read_assignment.assignment_type.is_unassigned() or not read_assignment.isoform_matches:
            self.not_assigned_reads += 1
            self.reads_for_tpm[group_id] += 1
            return
        elif read_assignment.isoform_matches and read_assignment.isoform_matches[0].assigned_transcript is None:
            self.not_assigned_reads += 1
            self.reads_for_tpm[group_id] += 1
            logger.warning("Assigned feature for read %s is None, will be skipped. "
                           "This message may be reported to the developers." % read_assignment.read_id)
            return

        feature_ids = self.assignment_extractor.get_features(read_assignment)
        assignment_type = self.assignment_extractor.get_assignment_type(read_assignment)
        self.reads_for_tpm[group_id] += 1
        if assignment_type == ReadAssignmentType.ambiguous:
            count_value = self.read_counter.process_ambiguous(len(feature_ids))
            for feature_id in feature_ids:
                self.feature_counter[feature_id].inc(group_id, count_value)
                if count_value > 0:
                    self.all_features.add(feature_id)
            self.ambiguous_reads += 1

        elif assignment_type.is_inconsistent():
            count_value = self.read_counter.process_inconsistent(assignment_type, len(feature_ids))
            if count_value > 0:
                for feature_id in feature_ids:
                    self.feature_counter[feature_id].inc(group_id, count_value)
                    self.all_features.add(feature_id)

        elif assignment_type.is_unique():
            feature_id = list(feature_ids)[0]
            self.feature_counter[feature_id].inc(group_id, 1.0)
            self.all_features.add(feature_id)
            if self.assignment_extractor.confirms_feature(read_assignment):
                self.confirmed_features.add(feature_id)

    def add_read_info_raw(self, read_id, feature_ids, group_ids):
        # group_ids is now a list of integers (pool indices)
        if self.ignore_read_groups or group_ids is None:
            group_id = 0  # Default group index
        else:
            group_id = group_ids[self.group_index]
        if not read_id:
            self.not_aligned_reads += 1
        elif not feature_ids:
            self.not_assigned_reads += 1
            self.reads_for_tpm[group_id] += 1
        elif len(feature_ids) > 1:
            self.ambiguous_reads += 1
            self.reads_for_tpm[group_id] += 1
            for feature_id in feature_ids:
                count_value = self.read_counter.process_ambiguous(len(feature_ids))
                self.feature_counter[feature_id].inc(group_id, count_value)
                self.all_features.add(feature_id)
        else:
            self.feature_counter[feature_ids[0]].inc(group_id, 1.0)
            self.all_features.add(feature_ids[0])
            self.reads_for_tpm[group_id] += 1

    def add_unassigned(self, read_assignment):
        # Use read_group_ids directly (integers)
        if self.ignore_read_groups:
            group_id = 0
        elif not read_assignment.read_group_ids:
            group_id = 0
        else:
            group_id = read_assignment.read_group_ids[self.group_index]
        self.not_assigned_reads += 1
        self.reads_for_tpm[group_id] += 1

    def add_unaligned(self, n_reads=1):
        self.not_aligned_reads += n_reads

    def format_header(self, all_groups, value_name="count"):
        if self.ignore_read_groups:
            return "feature_id\t%s\n" % value_name
        else:
            return "feature_id\t" + "\t".join(all_groups) + "\n"

    def add_confirmed_features(self, features):
        self.confirmed_features.update(features)

    def dump(self):
        all_features = sorted(filter(lambda x: x is not None, self.all_features))

        for feature_id in all_features:
            if feature_id in self.confirmed_features:
                continue
            # ignoring unconfirmed features
            for g in self.feature_counter[feature_id].data.keys():
                self.feature_counter[feature_id].data[g] = 0.0

        if self.ignore_read_groups:
            self.dump_ungrouped(all_features)
        else:
            self.dump_grouped(all_features)
        self.dump_usable()

    def _get_ordered_groups(self):
        """Get ordered list of group names from string pool."""
        if self.string_pools is None:
            return [AbstractReadGrouper.default_group_id]
        pool = self.string_pools.get_read_group_pool(self.group_index)
        return [pool.get_str(i) for i in range(len(pool))]

    def _get_group_name(self, group_id: int) -> str:
        """Get group name from pool for a group ID."""
        if self.string_pools is None:
            return AbstractReadGrouper.default_group_id
        pool = self.string_pools.get_read_group_pool(self.group_index)
        return pool.get_str(group_id)

    def _get_num_groups(self) -> int:
        """Get number of groups from pool."""
        if self.string_pools is None:
            return 1
        pool = self.string_pools.get_read_group_pool(self.group_index)
        return len(pool)

    def dump_ungrouped(self, all_features):
        with self.get_output_file_handler() as output_file:
            default_group_id = 0  # Default group index
            output_file.write(self.format_header(self._get_ordered_groups()))
            for feature_id in all_features:
                count = self.feature_counter[feature_id].get(default_group_id)
                output_file.write("%s\t%.2f\n" % (feature_id, count))

            with open(self.output_stats_file_name, "w") as f:
                f.write("__ambiguous\t%d\n" % self.ambiguous_reads)
                f.write("__no_feature\t%d\n" % self.not_assigned_reads)
                f.write("__not_aligned\t%d\n" % self.not_aligned_reads)

    def dump_grouped(self, all_features):
        with self.get_output_file_handler() as output_file:
            output_file.write("feature_id\tgroup_id\tcount\n")
            for feature_id in all_features:
                row_count = 0
                for group_id in self.feature_counter[feature_id].data.keys():
                    count = self.feature_counter[feature_id].data[group_id]
                    if count != 0:
                        group_name = self._get_group_name(group_id)
                        output_file.write("%s\t%s\t%.2f\n" % (feature_id, group_name, count))
                    row_count += count
            output_file.close()

    def dump_usable(self):
        with open(self.usable_file_name, "w") as f:
            # Iterate over all groups in pool order
            for group_id in range(self._get_num_groups()):
                group_name = self._get_group_name(group_id)
                f.write("%s\t%d\n" % (group_name, self.reads_for_tpm[group_id]))

    def load_usable(self, usable_file_name):
        if self.string_pools is None:
            # Ungrouped: just load into default group
            for l in open(usable_file_name, "r"):
                v = l.strip().split('\t')
                self.reads_for_tpm[0] += int(v[1])
        else:
            pool = self.string_pools.get_read_group_pool(self.group_index)
            for l in open(usable_file_name, "r"):
                v = l.strip().split('\t')
                group_name = v[0]
                if group_name in pool.str_to_int:
                    group_id = pool.str_to_int[group_name]
                    self.reads_for_tpm[group_id] += int(v[1])

    def finalize(self, args=None):
        if not args:
            normalization_method = NormalizationMethod.none
        elif args.normalization_method not in NormalizationMethod.__dict__:
            logger.warning("%s in not a valid normalization method, will use simple normalization" % args.normalization_method)
            normalization_method = NormalizationMethod.simple
        else:
            normalization_method = NormalizationMethod[args.normalization_method]

        if self.ignore_read_groups:
            if normalization_method == NormalizationMethod.none:
                return
            convert_ungrouped_to_tpm(self.output_file, self.output_tpm_file_name,
                                     normalization_method, self.reads_for_tpm[0])  # Default group index is 0
            return

        num_groups = self._get_num_groups()
        counts_format = {GroupedOutputFormat[f] for f in args.counts_format}
        explicit_matrix = GroupedOutputFormat.matrix in counts_format
        default_format = GroupedOutputFormat.default in counts_format
        actual_num_groups = num_groups
        if explicit_matrix or default_format:
            max_groups = GROUP_COUNT_CUTOFF if default_format and not explicit_matrix else 0
            # When pool count is reliable (string_pools available), skip without reading the file
            pool_reliable = self.string_pools is not None
            if pool_reliable and max_groups > 0 and num_groups > max_groups:
                logger.info("Skipping full matrix conversion: %d groups exceeds limit of %d. "
                            "Use '--counts_format matrix' to force conversion." % (num_groups, max_groups))
            else:
                actual_num_groups = convert_to_matrix(self.output_file, self.output_counts_prefix, max_groups=max_groups) or num_groups
                if normalization_method != NormalizationMethod.none:
                    reads_for_tpm = None
                    if normalization_method == NormalizationMethod.usable_reads:
                        reads_for_tpm = {self._get_group_name(group_id): self.reads_for_tpm[group_id]
                                         for group_id in range(num_groups)}
                    convert_to_matrix(self.output_file, self.output_tpm_prefix,
                                      convert_to_tpm=True, usable_reads_per_group=reads_for_tpm,
                                      max_groups=max_groups)

        if (GroupedOutputFormat.mtx in counts_format or
                (default_format and actual_num_groups > GROUP_COUNT_CUTOFF)):
            convert_to_mtx(self.output_file, self.output_counts_prefix)
            if normalization_method != NormalizationMethod.none:
                reads_for_tpm = None
                if normalization_method == NormalizationMethod.usable_reads:
                    reads_for_tpm = {self._get_group_name(group_id): self.reads_for_tpm[group_id]
                                     for group_id in range(num_groups)}
                convert_to_mtx(self.output_file, self.output_tpm_prefix,
                               convert_to_tpm=True, usable_reads_per_group=reads_for_tpm)

def create_gene_counter(output_file_name, strategy, complete_feature_list=None, string_pools=None, group_index: int = 0):
    read_weight_counter = ReadWeightCounter(strategy)
    return AssignedFeatureCounter(output_file_name, GeneAssignmentExtractor,
                                  string_pools, read_weight_counter, complete_feature_list, group_index)


def create_transcript_counter(output_file_name, strategy, complete_feature_list=None,
                              string_pools=None, group_index: int = 0):
    read_weight_counter = ReadWeightCounter(strategy)
    return AssignedFeatureCounter(output_file_name, TranscriptAssignmentExtractor,
                                  string_pools, read_weight_counter, complete_feature_list, group_index)


@unique
class NormalizationMethod(Enum):
    none = 0
    simple = 1
    usable_reads = 2


def convert_ungrouped_to_tpm(counts_file_name, output_tpm_file_name, normalization=NormalizationMethod.simple, reads_for_tpm=-1):
    total_counts = 0
    with open(counts_file_name, "r") as f:
        for line in f:
            if line.startswith('_'): continue
            if line.startswith('feature_id\t'): continue
            fs = line.rstrip().split('\t')
            total_counts += float(fs[1])

    if total_counts == 0:
        logger.error("No counts were detected in %s, cannot convert to TPM" % counts_file_name)
        return

    unassigned_tpm = 0.0
    if normalization == NormalizationMethod.usable_reads and reads_for_tpm > 0:
        total_reads = reads_for_tpm
        unassigned_tpm = 1000000.0 * (1 - total_counts / total_reads)
    else:
        total_reads = total_counts
    scale_factor = 1000000.0 / total_reads

    with open(output_tpm_file_name, "w") as outf:
        with open(counts_file_name) as f:
            for line in f:
                if line.startswith('_'): break
                if line.startswith('feature_id\t'):
                    outf.write(line.replace("count", "TPM"))
                    continue
                fs = line.rstrip().split('\t')
                feature_id, count = fs[0], float(fs[1])
                tpm = scale_factor * count
                outf.write("%s\t%.6f\n" % (feature_id, tpm))
            if unassigned_tpm > 0:
                outf.write("%s\t%.6f\n" % ("__unassigned", unassigned_tpm))


# count simple features inclusion/exclusion (exons / introns)
class ProfileFeatureCounter(AbstractCounter):
    def __init__(self, output_prefix, string_pools=None, group_index: int = 0):
        AbstractCounter.__init__(self, output_prefix, string_pools is None)
        self.string_pools = string_pools
        self.group_index = group_index  # Index in read_group list to use
        # feature_id -> (group_id -> count)  -- group_id is now an integer pool index
        self.inclusion_feature_counter = defaultdict(lambda: IncrementalDict(int))
        self.exclusion_feature_counter = defaultdict(lambda: IncrementalDict(int))
        self.feature_name_dict = OrderedDict()
        # Track which group IDs have been encountered (for dump)
        self.encountered_group_ids = set()

    def add_read_info_from_profile(self, gene_feature_profile, read_assigned_strand, feature_property_map,
                                   group_id: int = 0):
        """Process feature profile. group_id is an integer pool index."""
        self.encountered_group_ids.add(group_id)

        for i in range(len(gene_feature_profile)):
            if read_assigned_strand not in feature_property_map[i].strand:
                # skip features that do not match read's strand
                continue
            if gene_feature_profile[i] == 1:
                feature_id = feature_property_map[i].id
                self.inclusion_feature_counter[feature_id].inc(group_id)
                if feature_id not in self.feature_name_dict:
                    self.feature_name_dict[feature_id] = feature_property_map[i].to_str()
            elif gene_feature_profile[i] == -1:
                feature_id = feature_property_map[i].id
                self.exclusion_feature_counter[feature_id].inc(group_id)
                if feature_id not in self.feature_name_dict:
                    self.feature_name_dict[feature_id] = feature_property_map[i].to_str()

    def _get_group_name(self, group_id: int) -> str:
        """Get group name from pool for a group ID."""
        if self.string_pools is None:
            return AbstractReadGrouper.default_group_id
        pool = self.string_pools.get_read_group_pool(self.group_index)
        return pool.get_str(group_id)

    def dump(self):
        with open(self.output_counts_file_name, "w") as f:
            f.write(FeatureInfo.header() + "\tgroup_id\tinclude_counts\texclude_counts\n")

            all_group_ids = sorted(self.encountered_group_ids)
            for feature_id in self.feature_name_dict.keys():
                for group_id in all_group_ids:
                    feature_name = self.feature_name_dict[feature_id]
                    group_name = self._get_group_name(group_id)
                    incl_count = self.inclusion_feature_counter[feature_id].get(group_id)
                    excl_count = self.exclusion_feature_counter[feature_id].get(group_id)
                    if incl_count > 0 or excl_count > 0:
                        f.write("%s\t%s\t%d\t%d\n" % (feature_name, group_name, incl_count, excl_count))

    def finalize(self, args=None):
        pass

    @staticmethod
    def is_valid(assignment):
        return assignment is not None and \
               hasattr(assignment, 'exon_gene_profile') and assignment.exon_gene_profile is not None and \
               hasattr(assignment, 'intron_gene_profile') and assignment.intron_gene_profile is not None and \
               hasattr(assignment, 'gene_info') and assignment.gene_info is not None

    @staticmethod
    def is_assigned_to_gene(assignment):
        return not assignment.gene_assignment_type.is_unassigned() and not assignment.gene_assignment_type.is_ambiguous()


class ExonCounter(ProfileFeatureCounter):
    def __init__(self, output_prefix, string_pools=None, group_index: int = 0):
        ProfileFeatureCounter.__init__(self, output_prefix, string_pools, group_index)

    def add_read_info(self, read_assignment):
        if not ProfileFeatureCounter.is_valid(read_assignment) or not ProfileFeatureCounter.is_assigned_to_gene(read_assignment):
            return
        # Use read_group_ids directly (integers)
        if self.ignore_read_groups:
            group_id = 0
        elif not read_assignment.read_group_ids:
            group_id = 0
        else:
            group_id = read_assignment.read_group_ids[self.group_index]
        self.add_read_info_from_profile(read_assignment.exon_gene_profile, read_assignment.strand,
                                        read_assignment.gene_info.exon_property_map, group_id)


class IntronCounter(ProfileFeatureCounter):
    def __init__(self, output_prefix, string_pools=None, group_index: int = 0):
        ProfileFeatureCounter.__init__(self, output_prefix, string_pools, group_index)

    def add_read_info(self, read_assignment):
        if not ProfileFeatureCounter.is_valid(read_assignment) or not ProfileFeatureCounter.is_assigned_to_gene(read_assignment):
            return
        # Use read_group_ids directly (integers)
        if self.ignore_read_groups:
            group_id = 0
        elif not read_assignment.read_group_ids:
            group_id = 0
        else:
            group_id = read_assignment.read_group_ids[self.group_index]
        self.add_read_info_from_profile(read_assignment.intron_gene_profile, read_assignment.strand,
                                        read_assignment.gene_info.intron_property_map, group_id)




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
from collections import namedtuple
from enum import Enum, unique
from functools import partial

from .common import (
    contains,
    contains_approx,
    difference_in_present_features,
    equal_profiles_in_range,
    equal_ranges,
    extra_exon_percentage,
    has_overlapping_features,
    jaccard_similarity,
    overlap_intervals,
    overlaps,
    read_coverage_fraction,
    rindex,
)
from .isoform_assignment import (
    IsoformMatch,
    MatchEvent,
    MatchEventSubtype,
    MatchClassification,
    ReadAssignment,
    ReadAssignmentType,
    SupplementaryMatchConstants,
    event_subtype_cost,
    elongation_cost,
)
from .long_read_profiles import OverlappingFeaturesProfileConstructor
from .junction_comparator import JunctionComparator
from .polya_verification import PolyAVerifier


logger = logging.getLogger('IsoQuant')

IsoformDiff = namedtuple("IsoformDiff", ("id", "diff"))


@unique
class AmbiguityResolvingMethod(Enum):
    none = -10
    monoexon_only = 10
    monoexon_and_fsm = 20
    all = 30

    minimal_score = -0.5
    top_scored_factor = 1.5


class LongReadAssigner:
    def __init__(self, gene_info, params, string_pools, quick_mode=False):
        assert string_pools is not None, "string_pools is required"
        self.gene_info = gene_info
        self.params = params
        self.quick_mode = quick_mode
        self.string_pools = string_pools
        self.intron_comparator = JunctionComparator(params,
                                                    OverlappingFeaturesProfileConstructor
                                                    (self.gene_info.intron_profiles.features,
                                                     (self.gene_info.start, self.gene_info.end),
                                                     comparator=partial(equal_ranges, delta=self.params.delta)))
        self.polya_verifier = PolyAVerifier(self.gene_info, self.params)

    # ======== SUPPORT FUNCTIONS =======
    def get_gene_id(self, transcript_id):
        return self.gene_info.gene_id_map[transcript_id]

    @staticmethod
    def find_distances_to_pos(isoform_ids, transcript_function, pos):
        distances = []
        for isoform_id in isoform_ids:
            distances.append((abs(transcript_function(isoform_id) - pos), isoform_id))
        return distances

    @staticmethod
    def find_overlapping_isoforms(read_exon_split_profile, isoform_profiles, hint=None):
        isoforms = set()
        isoform_set = hint if hint is not None else isoform_profiles.profiles.keys()
        for isoform_id in isoform_set:
            profile_range = overlap_intervals(read_exon_split_profile.gene_profile_range,
                                              isoform_profiles.profile_ranges[isoform_id])
            if has_overlapping_features(isoform_profiles.profiles[isoform_id], read_exon_split_profile.gene_profile,
                                        profile_range=profile_range):
                isoforms.add(isoform_id)
        return isoforms

    def find_containing_isoforms(self, read_exon_split_profile, isoform_profiles, hint=None):
        isoforms = set()
        read_exons = read_exon_split_profile.read_features
        read_region = (read_exons[0][0], read_exons[-1][1])
        isoform_set = hint if hint is not None else isoform_profiles.keys()

        for isoform_id in isoform_set:
            if contains_approx(self.gene_info.transcript_region(isoform_id), read_region,
                               self.params.min_abs_exon_overlap):
                isoforms.add(isoform_id)
        return isoforms

    @staticmethod
    def match_profile(read_gene_profile, isoform_profiles, hint=None, diff_limit=-1, profile_range=None):
        """ match read profiles to a known isoform junction profile

        Parameters
        ----------
        read_gene_profile: list of int
        isoform_profiles: dict of str
        hint: set, optional
            potential candidates, all other profiles will be skipped

        Returns
        -------
        result: list of IsoformDiff
            list of tuples (id and difference)
        """
        isoforms = []
        isoform_set = hint if hint is not None else isoform_profiles.keys()
        for isoform_id in isoform_set:
            isoform_profile = isoform_profiles[isoform_id]
            diff = difference_in_present_features(isoform_profile, read_gene_profile, diff_limit, profile_range)
            if diff_limit == -1 or diff <= diff_limit:
                isoforms.append(IsoformDiff(isoform_id, diff))
        return sorted(isoforms, key=lambda x: x.diff)

    @staticmethod
    def match_profile_exact_in_range(read_gene_profile, isoform_profiles, profile_range, hint=None):
        isoforms = []
        isoform_set = hint if hint is not None else isoform_profiles.keys()
        for isoform_id in isoform_set:
            isoform_profile = isoform_profiles[isoform_id]
            if equal_profiles_in_range(isoform_profile, read_gene_profile, profile_range):
                isoforms.append(IsoformDiff(isoform_id, 0))
        return isoforms

    def find_matching_isoforms(self, read_profile, isoform_profiles, hint=None):
        """ match read profiles to a known isoform junction profile

        Parameters
        ----------
        read_gene_profile: list of int
        isoform_profiles: dict of str
        hint: set, optional
            potential candidates, all other profiles will be skipped

        Returns
        -------
        result: list(str)
            matching isoforms

        """
        isoforms = self.match_profile_exact_in_range(read_profile.gene_profile, isoform_profiles,
                                                     read_profile.gene_profile_range, hint)
        return [x[0] for x in isoforms]

    # select most similar isoform based on different criteria
    def select_similar_isoforms(self, combined_read_profile):
        read_split_exon_profile = combined_read_profile.read_split_exon_profile
        overlapping_isoforms = self.find_overlapping_isoforms(read_split_exon_profile,
                                                              self.gene_info.split_exon_profiles)
        if not overlapping_isoforms:
            return
        # select isoforms with non-negative nucleotide score
        significantly_overlapping_isoforms = \
            self.resolve_by_nucleotide_score(combined_read_profile, overlapping_isoforms,
                                             similarity_function=self.coverage_based_nucleotide_score,
                                             top_scored_factor=0)
        if not significantly_overlapping_isoforms:
            return

        # find intron matches
        intron_matching_isoforms = self.match_profile(combined_read_profile.read_intron_profile.gene_profile,
                                                      self.gene_info.intron_profiles.profiles,
                                                      profile_range=combined_read_profile.read_intron_profile.gene_profile_range,
                                                      hint=significantly_overlapping_isoforms)

        # add extra terminal bases as potential inconsistency event
        candidates = []
        read_region = (read_split_exon_profile.read_features[0][0], read_split_exon_profile.read_features[-1][1])
        for isoform_id, diff_introns in intron_matching_isoforms:
            transcript_start, transcript_end = self.gene_info.transcript_region(isoform_id)
            extra_left = 0.5 if read_region[0] + self.params.delta < transcript_start else 0
            extra_right = 0.5 if read_region[1] - self.params.delta > transcript_end else 0
            candidates.append((isoform_id, diff_introns + extra_right + extra_left))
        # select isoforms that have similar number of potential inconsistencies
        best_diff = min(candidates, key=lambda x: x[1])[1]
        best_candidates = [x[0] for x in filter(lambda x: x[1] <= best_diff + 3, candidates)]
        # logger.debug("+ + Closest matching isoforms " + str(best_candidates))

        return best_candidates

    # detect exon elongation subtyp
    def categorize_exon_elongation_subtype(self, read_split_exon_profile, isoform_id):
        split_exons = self.gene_info.split_exon_profiles.features
        isoform_profile = self.gene_info.split_exon_profiles.profiles[isoform_id]
        profile_range = self.gene_info.split_exon_profiles.profile_ranges[isoform_id]
        read_profile_range = read_split_exon_profile.gene_profile_range

        # find first and last common exons
        common_first_exon = -1
        isoform_first_exon = profile_range[0]
        for i in range(max(isoform_first_exon, read_profile_range[0]), len(split_exons)):
            if isoform_profile[i] == read_split_exon_profile.gene_profile[i] == 1:
                common_first_exon = i
                break

        common_last_exon = -1
        isoform_last_exon = profile_range[1] - 1
        for i in range(min(isoform_last_exon, read_profile_range[1] - 1), -1, -1):
            if isoform_profile[i] == read_split_exon_profile.gene_profile[i] == 1:
                common_last_exon = i
                break
        if common_first_exon == -1 or common_last_exon == -1:
            logger.warning(" + Odd case for exon elongation, no matching exons")

        events = []

        first_read_exon = read_split_exon_profile.read_features[0]
        if overlaps(first_read_exon, split_exons[common_first_exon]):
            extra_left = split_exons[common_first_exon][0] - first_read_exon[0]
            if common_first_exon == isoform_first_exon:
                # logger.debug("+ + Left: %d %d" % (common_first_exon, isoform_first_exon))
                if abs(extra_left) <= self.params.minor_exon_extension:
                    if abs(extra_left) <= self.params.delta:
                        left_match_event = MatchEventSubtype.terminal_site_match_left_precise
                    else:
                        left_match_event = MatchEventSubtype.terminal_site_match_left
                    events.append(MatchEvent(left_match_event, event_info=extra_left))

                if extra_left > self.params.minor_exon_extension:
                    events.append(MatchEvent(MatchEventSubtype.major_exon_elongation_left, event_info=extra_left))
                elif extra_left > self.params.delta:
                    events.append(MatchEvent(MatchEventSubtype.exon_elongation_left, event_info=extra_left))
            elif self.params.minor_exon_extension >= extra_left > self.params.delta:
                # minor exon elongations detected for non-terminal exons too
                events.append(MatchEvent(MatchEventSubtype.exon_elongation_left, event_info=extra_left))

        last_read_exon = read_split_exon_profile.read_features[-1]
        if overlaps(last_read_exon, split_exons[common_last_exon]):
            extra_right = last_read_exon[1] - split_exons[common_last_exon][1]
            if common_last_exon == isoform_last_exon:
                # logger.debug("+ + Right: %d %d" % (common_last_exon, isoform_last_exon))
                if  abs(extra_right) <= self.params.minor_exon_extension:
                    if abs(extra_right) <= self.params.delta:
                        right_match_event = MatchEventSubtype.terminal_site_match_right_precise
                    else:
                        right_match_event = MatchEventSubtype.terminal_site_match_right
                    events.append(MatchEvent(right_match_event, event_info=extra_right))

                if extra_right > self.params.minor_exon_extension:
                    events.append(MatchEvent(MatchEventSubtype.major_exon_elongation_right, event_info=extra_right))
                elif extra_right > self.params.delta:
                    events.append(MatchEvent(MatchEventSubtype.exon_elongation_right, event_info=extra_right))
            elif self.params.minor_exon_extension >= extra_right > self.params.delta:
                events.append(MatchEvent(MatchEventSubtype.exon_elongation_right, event_info=extra_right))

        return events

    # select best assignment based on nucleotide similarity
    # score = similarity_function(read_exons, isoform_exons)
    def resolve_by_nucleotide_score(self, combined_read_profile, matched_isoforms, similarity_function,
                                    min_similarity=AmbiguityResolvingMethod.minimal_score.value,
                                    top_scored_factor=AmbiguityResolvingMethod.top_scored_factor.value):
        if not matched_isoforms:
            return []

        # logger.debug("+ + + Resolving by nucleotide similarity")
        read_exons = combined_read_profile.read_split_exon_profile.read_features

        scores = []
        for isoform_id in matched_isoforms:
            isoform_exons = self.gene_info.all_isoforms_exons[isoform_id]
            scores.append((isoform_id, similarity_function(read_exons, isoform_exons)))

        #logger.debug("Scores: " + str(scores))
        scores = sorted(scores, key=lambda x:x[1], reverse=True)
        best_score = scores[0][1]
        if top_scored_factor == 0:
            top_scored = sorted(filter(lambda x: x[1] >= min_similarity, scores))
        else:
            top_scored = sorted(filter(lambda x: x[1] * top_scored_factor >= best_score and x[1] >= min_similarity,
                                       scores))
        # logger.debug("+ + + Best score = %f, all candidates = %s" % (best_score, str(top_scored)))
        return list(map(lambda x: x[0], top_scored))

    # score = read's covered fraction - flanking fraction (can take values in [-1,1])
    # gives priority to isoforms that contain the read
    def coverage_based_nucleotide_score(self, read_exons, isoform_exons):
        read_coverage = read_coverage_fraction(read_exons, isoform_exons)
        isoform_extended_region = (isoform_exons[0][0] - self.params.minor_exon_extension,
                                   isoform_exons[-1][1] + self.params.minor_exon_extension)
        flanking_percentage = extra_exon_percentage(isoform_extended_region, read_exons)

        return read_coverage - flanking_percentage

    # score = Jaccard similarity - flanking fraction (can take values in [-1,1])
    # gives priority to isoforms that contain the read
    def jaccard_based_nucleotide_score(self, read_exons, isoform_exons):
        js = jaccard_similarity(read_exons, isoform_exons)
        isoform_extended_region = (isoform_exons[0][0] - self.params.minor_exon_extension,
                                   isoform_exons[-1][1] + self.params.minor_exon_extension)
        flanking_percentage = extra_exon_percentage(isoform_extended_region, read_exons)

        return js - flanking_percentage

    # ====== CLASSIFICATION =======

    # check for extra sequences and modify assignment accordingly
    def check_read_ends(self, read_split_exon_profile, assignment):
        for match in assignment.isoform_matches:
            if match.assigned_transcript is None:
                continue

            exon_elongation_types = self.categorize_exon_elongation_subtype(read_split_exon_profile,
                                                                            match.assigned_transcript)
            for e in exon_elongation_types:
                match.add_subclassification(e)

            if any(MatchEventSubtype.is_major_elongation(e.event_type) for e in exon_elongation_types):
                # serious exon elongation
                if not assignment.assignment_type.is_inconsistent():
                    assignment.set_assignment_type(ReadAssignmentType.inconsistent_non_intronic)
            elif any(MatchEventSubtype.is_minor_elongation(e.event_type) for e in exon_elongation_types):
                # minor exon elongation
                if assignment.assignment_type == ReadAssignmentType.unique:
                    assignment.set_assignment_type(ReadAssignmentType.unique_minor_difference)

    # get incompleteness type, assuming introns are matched and read is spliced
    def detect_ism_subtype(self, read_region, isoform_id):
        isoform_introns = self.gene_info.all_isoforms_introns[isoform_id]
        isoform_intron_region = (isoform_introns[0][0], isoform_introns[-1][1])
        is_left_truncated = isoform_intron_region[0] < read_region[0]
        is_right_truncated = isoform_intron_region[1] > read_region[1]

        if is_left_truncated and is_right_truncated:
            # logger.debug(" + Internal")
            event_type = MatchEventSubtype.ism_internal
        elif is_left_truncated:
            # logger.debug(" + Truncated on the left")
            event_type = MatchEventSubtype.ism_left
        elif is_right_truncated:
            # logger.debug(" + Truncated on the right")
            event_type = MatchEventSubtype.ism_right
        else:
            # logger.debug(" + No ISM truncation ")
            event_type = MatchEventSubtype.none
        return MatchEvent(event_type)

    # check where it is full splice match, assume all introns are matched and read is spliced
    def is_fsm(self, read_region, isoform_id):
        isoform_introns = self.gene_info.all_isoforms_introns[isoform_id]
        # all isoform instrons are within read
        return contains(read_region, (isoform_introns[0][0], isoform_introns[-1][1]))

    # make proper match subtype, assume all introns are matched or read is unspliced
    def categorize_correct_splice_match(self, combined_read_profile, isoform_id):
        read_exons = combined_read_profile.read_split_exon_profile.read_features
        read_region = (read_exons[0][0], read_exons[-1][1])

        if len(combined_read_profile.read_intron_profile.read_profile) == 0 or \
                len(self.gene_info.all_isoforms_introns[isoform_id]) == 0:
            # logger.debug(" + Mono exon")
            match_classification = MatchClassification.mono_exon_match
            match_subclassifications = MatchEvent(MatchEventSubtype.mono_exon_match)
        elif self.is_fsm(read_region, isoform_id):
            # logger.debug("+ + Full splice match " + isoform_id)
            match_classification = MatchClassification.full_splice_match
            match_subclassifications = MatchEvent(MatchEventSubtype.fsm)
        else:
            # logger.debug("+ + Incomplete splice match " + isoform_id)
            match_classification = MatchClassification.incomplete_splice_match
            match_subclassifications = self.detect_ism_subtype(read_region, isoform_id)

        return IsoformMatch(match_classification, self.string_pools,
                            assigned_gene=self.get_gene_id(isoform_id),
                            assigned_transcript=isoform_id,
                            match_subclassification=match_subclassifications,
                            transcript_strand=self.gene_info.isoform_strands[isoform_id])

    # make proper match subtype
    def categorize_multiple_splice_matches(self, combined_read_profile, isoform_ids):
        isoform_matches = []
        for isoform_id in sorted(isoform_ids):
            isoform_matches.append(self.categorize_correct_splice_match(combined_read_profile, isoform_id))
        return isoform_matches

    def categorize_correct_unspliced_match(self, combined_read_profile, isoform_id):
        isoform_exons = self.gene_info.all_isoforms_exons[isoform_id]
        if len(isoform_exons) == 1:
            # both isoform and read are monoexon
            events = [MatchEvent(MatchEventSubtype.mono_exon_match)]
        else:
            # monoexonic read is inside exon
            events = [MatchEvent(MatchEventSubtype.mono_exonic)]

        match_classification = MatchClassification.get_mono_exon_classification(events)
        return IsoformMatch(match_classification, self.string_pools,
                            assigned_gene=self.get_gene_id(isoform_id),
                            assigned_transcript=isoform_id,
                            match_subclassification=events,
                            transcript_strand=self.gene_info.isoform_strands[isoform_id])

    def categorize_multiple_unspliced_matches(self, combined_read_profile, isoform_ids):
        isoform_matches = []
        for isoform_id in sorted(isoform_ids):
            isoform_matches.append(self.categorize_correct_unspliced_match(combined_read_profile, isoform_id))
        return isoform_matches

    # =========== MAIN PART ============

    # === Isoform matching function ===
    def assign_to_isoform(self, read_id, combined_read_profile):
        """ assign read to isoform according to it

        Parameters
        ----------
        read_id: str
        combined_read_profile: CombinedProfile

        Returns
        -------

        """
        read_intron_profile = combined_read_profile.read_intron_profile
        read_split_exon_profile = combined_read_profile.read_split_exon_profile

        # logger.debug("> Read blocks" + str(read_split_exon_profile.read_features))
        # logger.debug("Gene split exons" + str(self.gene_info.split_exon_profiles.features))
        # logger.debug("Read exons profile" + str(read_split_exon_profile.read_profile))
        # logger.debug("Gene exons profile" + str(read_split_exon_profile.gene_profile))

        # logger.debug("< Read introns" + str(read_intron_profile.read_features))
        # logger.debug("Gene introns" + str(self.gene_info.intron_profiles.features))
        # logger.debug("Read intron profile" + str(read_intron_profile.read_profile))
        # logger.debug("Gene intron profile" + str(read_intron_profile.gene_profile))
        if self.gene_info.empty():
            return ReadAssignment(read_id, ReadAssignmentType.intergenic, self.string_pools,
                                  match=IsoformMatch(MatchClassification.intergenic, string_pools=self.string_pools))

        if all(el != 1 for el in read_split_exon_profile.read_profile) \
                or all(el == 0 or el == -2 for el in read_split_exon_profile.gene_profile):
            read_region = (read_split_exon_profile.read_features[0][0], read_split_exon_profile.read_features[-1][1])
            gene_region = (self.gene_info.split_exon_profiles.features[0][0],
                              self.gene_info.split_exon_profiles.features[-1][1])
            # none of the blocks matched
            if not overlaps(read_region, gene_region):
                # logger.debug("EMPTY - noninformative")
                assignment = ReadAssignment(read_id, ReadAssignmentType.noninformative, self.string_pools,
                                          match=IsoformMatch(MatchClassification.intergenic, string_pools=self.string_pools))
            elif all(el != 1 for el in read_split_exon_profile.gene_profile):
                # logger.debug("EMPTY - intronic")
                # TODO: match to a gene
                assignment = ReadAssignment(read_id, ReadAssignmentType.noninformative, self.string_pools,
                                          match=IsoformMatch(MatchClassification.genic_intron, string_pools=self.string_pools))
            else:
                # logger.debug("EMPTY - genic")
                assignment = ReadAssignment(read_id, ReadAssignmentType.noninformative, self.string_pools,
                                          match=IsoformMatch(MatchClassification.genic, string_pools=self.string_pools))
            return assignment

        elif any(el == -1 for el in read_intron_profile.read_profile) \
                or any(el == -1 for el in read_split_exon_profile.read_profile):
            # Read has inconsistent exons / introns
            # logger.debug("+ Has inconsistent features (novel intron / exons)")
            assignment = self.match_inconsistent(read_id, combined_read_profile)

        elif any(el == 0 for el in read_intron_profile.read_profile) \
                or any(el == 0 for el in read_split_exon_profile.read_profile):
            # Read has extra flanking exons / introns
            # logger.debug("+ Has extra flanking features")
            assignment = self.match_inconsistent(read_id, combined_read_profile)

        else:
            # logger.debug("+ No contradictory features in read, but inconsistent isoforms may still be detected")
            assignment = self.match_consistent(read_id, combined_read_profile)

            if assignment is None:
                # alternative isoforms made of known introns/exons or intron retention
                # logger.debug("+ + Resolving unmatched ")
                assignment = self.match_inconsistent(read_id, combined_read_profile)

        return assignment

    def match_consistent(self, read_id, combined_read_profile):
        """ match profile when all read features are assigned and read has at least one intron

        Parameters
        ----------
        read_id: str
        combined_read_profile: CombinedReadProfiles

        Returns
        -------

        """
        read_intron_profile = combined_read_profile.read_intron_profile
        read_split_exon_profile = combined_read_profile.read_split_exon_profile
        isoform_split_exon_profiles = self.gene_info.split_exon_profiles.profiles

        containing_isoforms = self.find_containing_isoforms(read_split_exon_profile, isoform_split_exon_profiles)
        if not containing_isoforms:
            return None

        overlapping_isoforms = self.find_overlapping_isoforms(read_split_exon_profile, self.gene_info.split_exon_profiles,
                                                              hint=containing_isoforms)
        if not overlapping_isoforms:
            return None

        consistent_isoforms = self.find_matching_isoforms(read_intron_profile,
                                                          self.gene_info.intron_profiles.profiles,
                                                          hint=overlapping_isoforms)

        if not read_intron_profile.read_profile:
            assignment = self.match_consistent_unspliced(read_id, combined_read_profile, consistent_isoforms)
        else:
            assignment = self.match_consistent_spliced(read_id, combined_read_profile, consistent_isoforms)

        if assignment is not None:
            self.check_read_ends(combined_read_profile.read_split_exon_profile, assignment)
            self.verify_read_ends_for_assignment(combined_read_profile, assignment)

            # if apa detected isoform should be treated as inconsistent and rechecked
            if assignment.assignment_type.is_inconsistent():
                return None

        return assignment

    def match_consistent_spliced(self, read_id, combined_read_profile, consistent_isoforms):
        isoform_split_exon_profiles = self.gene_info.split_exon_profiles.profiles
        matched_isoforms = consistent_isoforms
        if len(consistent_isoforms) > 1:
            exon_matched_isoforms = \
                self.find_matching_isoforms(combined_read_profile.read_split_exon_profile,
                                            isoform_split_exon_profiles, hint=consistent_isoforms)
            if len(exon_matched_isoforms) != 0:
                matched_isoforms = exon_matched_isoforms

        if len(matched_isoforms) > 1:
            read_exons = combined_read_profile.read_split_exon_profile.read_features
            read_region = (read_exons[0][0], read_exons[-1][1])
            if self.params.resolve_ambiguous == AmbiguityResolvingMethod.all or \
               (self.params.resolve_ambiguous == AmbiguityResolvingMethod.monoexon_and_fsm and
                any(self.is_fsm(read_region, isoform_id) for isoform_id in matched_isoforms)):
                # resolve spliced using nucleotide score
                matched_isoforms = \
                    self.resolve_by_nucleotide_score(combined_read_profile, matched_isoforms,
                                                     similarity_function=self.jaccard_based_nucleotide_score)

        read_assignment = None
        if len(matched_isoforms) == 1:
            isoform_id = list(matched_isoforms)[0]
            # logger.debug("+ + UNIQUE intron match found " + isoform_id)
            isoform_match = self.categorize_correct_splice_match(combined_read_profile, isoform_id)
            read_assignment = ReadAssignment(read_id, ReadAssignmentType.unique, self.string_pools,
                                            match=isoform_match)

        elif len(matched_isoforms) > 1:
            # logger.debug("+ + Ambiguous read")
            isoform_matches = self.categorize_multiple_splice_matches(combined_read_profile, matched_isoforms)
            return ReadAssignment(read_id, ReadAssignmentType.ambiguous, self.string_pools,
                                match=isoform_matches)

        return read_assignment

    def match_consistent_unspliced(self, read_id, combined_read_profile, consistent_isoforms):
        # logger.debug("+  Resolving monoexonic read")

        matched_isoforms = consistent_isoforms
        if len(consistent_isoforms) > 1  and self.params.resolve_ambiguous != AmbiguityResolvingMethod.none:
            matched_isoforms = self.resolve_by_nucleotide_score(combined_read_profile, sorted(consistent_isoforms),
                                                                similarity_function=self.jaccard_based_nucleotide_score)

        read_assignment = None
        if len(matched_isoforms) == 1:
            isoform_id = matched_isoforms[0]
            # logger.debug("Single monoexonic match: %s" % isoform_id)
            isoform_match = self.categorize_correct_unspliced_match(combined_read_profile, isoform_id)
            read_assignment = ReadAssignment(read_id, ReadAssignmentType.unique, self.string_pools,
                                            match=isoform_match)

        elif len(matched_isoforms) > 1:
            # logger.debug("Nucleotide similarity picked multiple isoforms")
            isoform_matches = self.categorize_multiple_unspliced_matches(combined_read_profile, matched_isoforms)
            read_assignment = ReadAssignment(read_id, ReadAssignmentType.ambiguous, self.string_pools,
                                            match=isoform_matches)

        return read_assignment

    # resolve when there are -1s in read profile or when there are no exactly matching isoforms, but no -1s in read profiles
    def match_inconsistent(self, read_id, combined_read_profile):
        if self.quick_mode:
            return ReadAssignment(read_id, ReadAssignmentType.inconsistent, self.string_pools,
                                match=IsoformMatch(MatchClassification.genic, string_pools=self.string_pools))

        # select most similar isoforms based on multiple criteria
        best_candidates = self.select_similar_isoforms(combined_read_profile)
        if not best_candidates:
            return ReadAssignment(read_id, ReadAssignmentType.noninformative, self.string_pools,
                                match=IsoformMatch(MatchClassification.genic, string_pools=self.string_pools))

        # logger.debug("* Best candidates for inconsistency detection: " + str(best_candidates))
        # detect inconsistency for each one
        read_matches = self.detect_inconsistensies(read_id, combined_read_profile, best_candidates)
        if not read_matches:
            return ReadAssignment(read_id, ReadAssignmentType.noninformative, self.string_pools)
        # logger.debug("* Inconsistencies detected: " + str(read_matches))

        # select ones with the least number of inconsistent events
        best_isoforms, penalty_score = self.select_best_among_inconsistent(combined_read_profile, read_matches)
        if not best_isoforms:
            return ReadAssignment(read_id, ReadAssignmentType.noninformative, self.string_pools)
        # logger.debug("* Selected isoforms: " + str(best_isoforms))

        best_isoforms = sorted(best_isoforms)
        assignment_type = self.classify_assignment(best_isoforms, read_matches)
        if not combined_read_profile.read_intron_profile.read_profile:
            isoform_matches = self.create_monoexon_matches(read_matches, best_isoforms)
        elif assignment_type.is_inconsistent():
            isoform_matches = self.create_inconsistent_matches(read_matches, best_isoforms, penalty_score)
        else:
            isoform_matches = self.create_consistent_matches(read_matches, best_isoforms, combined_read_profile)
        return ReadAssignment(read_id, assignment_type, self.string_pools, match=isoform_matches)

    def classify_assignment(self, best_isoforms, read_matches):
        is_abmiguous = len(best_isoforms) > 1
        all_event_types = set()
        for isoform_id in best_isoforms:
            for e in read_matches[isoform_id]:
                all_event_types.add(e.event_type)

        #logger.debug("* All events: " + str(all_event_types))
        if all(MatchEventSubtype.is_consistent(e) for e in all_event_types):
            #logger.debug("* * Assignment seems to be consistent")
            assignment_type = ReadAssignmentType.ambiguous if is_abmiguous else ReadAssignmentType.unique
        elif any(MatchEventSubtype.is_major_inconsistency(e) for e in all_event_types):
            #logger.debug("* * Assignment is inconsistent")
            if any(MatchEventSubtype.is_intronic_inconsistency(e) for e in all_event_types):
                inconsistency_type = ReadAssignmentType.inconsistent
            else:
                inconsistency_type = ReadAssignmentType.inconsistent_non_intronic
            assignment_type = ReadAssignmentType.inconsistent_ambiguous if is_abmiguous else inconsistency_type
        elif any(MatchEventSubtype.is_minor_error(e) for e in all_event_types):
            #logger.debug("* * Assignment has minor errors")
            assignment_type = ReadAssignmentType.ambiguous if is_abmiguous else ReadAssignmentType.unique_minor_difference
        else:
            logger.warning("Unexpected event reported: " + str(all_event_types))
            assignment_type = ReadAssignmentType.noninformative

        return assignment_type

    def create_monoexon_matches(self, read_matches, selected_isoforms):
        matches = []
        for isoform_id in selected_isoforms:
            read_match = read_matches[isoform_id]
            match_classification = MatchClassification.get_mono_exon_classification(read_match)
            isoform_match = IsoformMatch(match_classification, self.string_pools,
                                         assigned_gene=self.get_gene_id(isoform_id),
                                         assigned_transcript=isoform_id,
                                         match_subclassification=read_match,
                                         transcript_strand=self.gene_info.isoform_strands[isoform_id])
            matches.append(isoform_match)
        return matches

    def create_consistent_matches(self, read_matches, selected_isoforms, combined_read_profile):
        matches = []
        for isoform_id in selected_isoforms:
            isoform_match = self.categorize_correct_splice_match(combined_read_profile, isoform_id)
            for e in read_matches[isoform_id]:
                if e.event_type != MatchEventSubtype.none:
                    isoform_match.add_subclassification(e)
            matches.append(isoform_match)
        return matches

    def create_inconsistent_matches(self, read_matches, selected_isoforms, penalty_score=0.0):
        matches = []
        for isoform_id in selected_isoforms:
            read_match = read_matches[isoform_id]
            match_classification = MatchClassification.get_inconsistency_classification(read_match)
            isoform_match = IsoformMatch(match_classification, self.string_pools,
                                         assigned_gene=self.get_gene_id(isoform_id),
                                         assigned_transcript=isoform_id,
                                         match_subclassification=read_match,
                                         transcript_strand=self.gene_info.isoform_strands[isoform_id],
                                         penalty_score=penalty_score)
            matches.append(isoform_match)
        return matches

    # detect inconsistensies for selected isoforms
    def detect_inconsistensies(self, read_id, combined_read_profile, matched_isoforms):
        """ compare read to closest matching isoforms

        Parameters
        ----------
        read_id: str
        matched_isoforms: list of str

        Returns
        -------
        result: list of tuples (isoform_id, list of events)
        """
        # get isoforms that have closes intron and exon profiles
        #logger.debug("Detecting difference for %s, %d matched isoforms" % (read_id, len(matched_isoforms)))
        read_intron_profile = combined_read_profile.read_intron_profile
        read_split_exon_profile = combined_read_profile.read_split_exon_profile

        #logger.debug("+ + Closest matching isoforms " + str(matched_isoforms))
        # isoform_id = best_isoform_ids[0]
        # logger.debug(str(self.gene_info.split_exon_profiles.profiles[isoform_id]))
        # logger.debug(str(self.gene_info.intron_profiles.profiles[isoform_id]))

        read_matches = {}
        for isoform_id in sorted(matched_isoforms):
            # logger.debug("Checking isoform %s" % isoform_id)
            # get intron coordinates
            isoform_introns = self.gene_info.all_isoforms_introns[isoform_id]
            # read start-end coordinates
            read_region = (read_split_exon_profile.read_features[0][0], read_split_exon_profile.read_features[-1][1])
            # isoform start-end
            isoform_region = self.gene_info.transcript_region(isoform_id)
            # logger.debug("R: " + str(combined_read_profile.read_split_exon_profile.read_features))
            # logger.debug(str(combined_read_profile.read_intron_profile.read_features))
            # logger.debug("I: " + str(self.gene_info.all_isoforms_exons[isoform_id]))

            matching_events = self.intron_comparator.compare_junctions(read_intron_profile.read_features, read_region,
                                                                       isoform_introns, isoform_region)

            if len(matching_events) == 1 and matching_events[0].event_type == MatchEventSubtype.undefined:
                continue

            matching_events += self.categorize_exon_elongation_subtype(read_split_exon_profile, isoform_id)
            matching_events = self.polya_verifier.verify_read_ends(combined_read_profile, isoform_id, matching_events)
            read_matches[isoform_id] = matching_events

            # logger.debug("+ + Found contradiction for " + isoform_id + ": " + " ".join(map(lambda x: x.event_type.name, matching_events)))
        return read_matches

    def select_best_among_inconsistent(self, combined_read_profile, read_matches):
        isoform_scores = []
        for isoform_id, match_events in read_matches.items():
            # logger.debug("* * Scoring inconsistencies for " + isoform_id + ": " + str(match_events))
            penalty_score = 0.0
            for e in match_events:
                event_count = 1
                if e.isoform_region != SupplementaryMatchConstants.undefined_region and \
                        SupplementaryMatchConstants.absent_position not in e.isoform_region and \
                        e.event_type in {MatchEventSubtype.exon_skipping_known, MatchEventSubtype.exon_skipping_novel}:
                    event_count = e.isoform_region[1] - e.isoform_region[0] + 1
                elif e.read_region != SupplementaryMatchConstants.undefined_region and \
                        SupplementaryMatchConstants.absent_position not in e.read_region:
                    event_count = e.read_region[1] - e.read_region[0] + 1
                    if e.event_type in {MatchEventSubtype.exon_gain_novel,
                                        MatchEventSubtype.exon_gain_known,
                                        MatchEventSubtype.mutually_exclusive_exons_novel,
                                        MatchEventSubtype.mutually_exclusive_exons_known,
                                        MatchEventSubtype.exon_detach_known,
                                        MatchEventSubtype.exon_detach_novel}:
                        event_count -= 1
                    elif e.event_type in {MatchEventSubtype.intron_retention,
                                          MatchEventSubtype.unspliced_intron_retention,
                                          MatchEventSubtype.fake_micro_intron_retention,
                                          MatchEventSubtype.incomplete_intron_retention_left,
                                          MatchEventSubtype.incomplete_intron_retention_right}:
                        event_count = 1

                event_cost = event_subtype_cost[e.event_type]
                if e.event_type in {MatchEventSubtype.major_exon_elongation_left,
                                    MatchEventSubtype.major_exon_elongation_right,
                                    MatchEventSubtype.exon_elongation_right,
                                    MatchEventSubtype.exon_elongation_left}:
                    event_cost = elongation_cost(self.params, e.event_info)

                penalty_score +=  event_cost * event_count
                # logger.debug("* * * Event " + str(e.event_type) + ", introns affected " + str(event_count) +
                #             ", event cost " + str(event_cost) +
                #             ". Updated penalty_score: " + str(penalty_score))
            # logger.debug("* * Final penalty_score for isoform " + isoform_id + ": " + str(penalty_score))
            isoform_scores.append((isoform_id, penalty_score))

        min_penalty_score = min(isoform_scores, key=lambda x:x[1])[1]
        # logger.debug("* * Best penalty_score " + str(min_penalty_score))
        best_isoforms = [x[0] for x in filter(lambda x:x[1] == min_penalty_score, isoform_scores)]
        # logger.debug("* * Best isoforms " + str(best_isoforms))

        # if several isoforms are tied select the best according to nucl penalty_score
        if len(best_isoforms) > 1:
            best_isoforms = self.resolve_by_nucleotide_score(combined_read_profile, best_isoforms,
                                                             similarity_function=self.coverage_based_nucleotide_score)
            # logger.debug("* * Resolved by nucl penalty_score " + str(best_isoforms))

        return best_isoforms, min_penalty_score

    # ==== POLYA STUFF ====
    def verify_read_ends_for_assignment(self, combined_read_profile, assignment):
        match_dict = {}
        for match in assignment.isoform_matches:
            if match.assigned_transcript is None:
                continue
            match.match_subclassifications = \
                self.polya_verifier.verify_read_ends(combined_read_profile, match.assigned_transcript,
                                                     match.match_subclassifications)
            match_dict[match.assigned_transcript] = match.match_subclassifications

        assignment.assignment_type = self.classify_assignment(match_dict.keys(), match_dict)



